import dataclasses
import datetime
import hashlib
import json
import os
import secrets
import tempfile
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import Iterator
from typing import Optional
from typing import Union
from urllib.parse import urlparse

import boto3
import pandas

from tecton_athena.templates_utils import load_template
from tecton_core import conf
from tecton_core import logger as logger_lib
from tecton_core.errors import TectonAthenaValidationError
from tecton_spark import offline_store

logger = logger_lib.get_logger("AthenaSession")

# In some cases, strings in Pandas DF are actually represented as "object" types.
# Hence the sketchy 'object' -> 'string' map
PANDAS_TO_HIVE_TYPES = {
    "string": "string",
    "object": "string",
    "int64": "bigint",
    "float64": "double",
    "int32": "bigint",
}

S3_ATHENA_PANDAS_UPLOADS = "athena/pandas_uploads"

CREATE_TABLE_TEMPLATE = load_template("create_table.sql")


PARTITION_TYPE_DATESTR = "PARTITION_TYPE_DATESTR"
PARTITION_TYPE_UNIX_EPOCH_NS = "PARTITION_TYPE_UNIX_EPOCH_NS"
GLUE_CATALOG_TABLE_PROPERTY_HASH = "tecton_table_metadata_hash"
GLUE_CATALOG_TABLE_PROPERTY_SPEC_VERSION = "tecton_table_spec_version"


@dataclass
class AthenaTableCreationSpec:
    """Athena Table representation (registered in Glue)"""

    # SPEC_VERSION Explanation:
    # - The SDK will never overwrite a table registration that happened with a newer version
    # - Increment if you fundamentally change the way Athena tables, for an identical FeatureView, are registered
    #   For example, if you move away from partition projection
    # - The SPEC_VERSION must not be part of the stable_hash calculation - otherwise the SDK believes it needs to overwrite
    #   a table's registration, even if the metadata for that table is unchanged, just because the version is incremented
    # - A separate SPEC_VERSION is used here rather than relying on the SDK's version because most SDK versions will increment
    #   without ever incrementing the SPEC_VERSION here
    SPEC_VERSION = 1

    database: str
    table: str
    s3_location: str
    data_format: str
    columns: Dict[str, str]
    partition_by: Union[offline_store.TIME_PARTITION, offline_store.ANCHOR_TIME]
    partition_by_type: str
    partition_by_format: str
    partition_by_range_from: str
    partition_by_range_to: str
    partition_by_interval: int
    partition_by_interval_timedelta: datetime.timedelta
    partition_type: Union[PARTITION_TYPE_DATESTR, PARTITION_TYPE_UNIX_EPOCH_NS]

    @property
    def _json_str(self):
        return json.dumps(
            dataclasses.asdict(self),
            default=str,
            sort_keys=True,
            indent=None,
            ensure_ascii=False,
            separators=(",", ":"),
        ).encode("utf-8")

    @property
    def stable_hash(self):
        str_representation = self._json_str
        return hashlib.md5(str_representation).hexdigest()


@dataclass
class AthenaSessionConfig:
    boto3_session: Optional[boto3.Session] = None
    workgroup: Optional[str] = None  # Athena workgroup.
    encryption: Optional[str] = None  # Valid values: [None, 'SSE_S3', 'SSE_KMS']. Notice: 'CSE_KMS' is not supported.
    kms_key: Optional[str] = None  # For SSE-KMS, this is the KMS key ARN or ID.
    enable_experimental_delta_support: bool = False  # Enables experimental Delta support. Note this will return incorrect values if you use optimize or delete.
    database: Optional[
        str
    ] = None  # Name of the Database in the Catalog. Can also be set via env variable ATHENA_DATABASE (backward compatibility)
    s3_path: Optional[
        str
    ] = None  # S3 Location where spines will be uploaded and Athena output will be stored. Can be set via env variable ATHENA_S3_PATH (backward compatibility)


class AthenaSession:
    def __init__(self):
        self._athena_s3_path = None
        self.config = AthenaSessionConfig()

    @property
    def _wr(self):
        try:
            import awswrangler

            return awswrangler
        except ModuleNotFoundError:
            raise Exception(
                "Athena Session cannot be initialized. Python module awswrangler not found. Did you forget to pip install tecton[athena]?"
            )

    def _get_athena_s3_path(self):
        s3_path = self.config.s3_path or conf.get_or_none("ATHENA_S3_PATH")
        if s3_path is not None:
            # Configuration always takes precedent and can be set at any time
            self._athena_s3_path = s3_path
        elif self._athena_s3_path is None:
            # If the bucket hasn't been initialized yet, let's create a new bucket
            # Let's cache the result to ensure we don't do it unnecessarily over and over again
            self._athena_s3_path = self._wr.athena.create_athena_bucket(boto3_session=self.config.boto3_session)

        if self._athena_s3_path.endswith("/"):
            # Drop "/" - calling function expects a path without trailing "/"
            self._athena_s3_path = self._athena_s3_path[0:-1]

        if not self._athena_s3_path.startswith("s3"):
            raise TectonAthenaValidationError(f"Provided S3 path does not start with s3://. Provided value: {s3_path}")

        return self._athena_s3_path

    def delete_table_if_exists(self, database: str, table: str):
        return self._wr.catalog.delete_table_if_exists(
            database=database, table=table, boto3_session=self.config.boto3_session
        )

    def delete_view_if_exists(self, view_name: str):
        sql = f"DROP VIEW IF EXISTS {view_name}"
        self.sql(sql)

    def _get_table_parameters(self, database: str, table: str) -> Dict:
        return self._wr.catalog.get_table_parameters(
            database=database, table=table, boto3_session=self.config.boto3_session
        )

    def _does_table_exist(self, database: str, table: str) -> bool:
        return self._wr.catalog.does_table_exist(
            database=database, table=table, boto3_session=self.config.boto3_session
        )

    def _upload_pandas_to_s3(self, pandas_df: pandas.DataFrame):
        with tempfile.NamedTemporaryFile(suffix=".parquet.snappy") as f:
            s3_client = boto3.client("s3")
            pandas_df.to_parquet(f.name, compression="snappy")

            local_file_name = os.path.basename(f.name)
            s3_random_dir_name = secrets.token_hex(10)

            s3_sub_directory = "/".join([S3_ATHENA_PANDAS_UPLOADS, s3_random_dir_name])

            s3_path_without_file = "/".join([self._get_athena_s3_path(), s3_sub_directory])
            s3_path_with_file = "/".join([s3_path_without_file, local_file_name])

            logger.info(f"Writing pandas df to S3 at {s3_path_with_file}...")

            # Consider switching to multipart upload
            # https://boto3.amazonaws.com/v1/documentation/api/latest/guide/s3.html#multipart-transfers
            s3_bucket = urlparse(s3_path_with_file).netloc

            # Skip leading slash
            s3_object_key = urlparse(s3_path_with_file).path[1:]
            # When kms key is provided, use it to encrypt the file
            extra_args = (
                {"ServerSideEncryption": "aws:kms", "SSEKMSKeyId": self.config.kms_key} if self.config.kms_key else {}
            )
            # Some customers need to use bucket, but others may not have permissions to do so.
            # Fallback to use client when permission is not there.
            try:
                s3 = self.config.boto3_session.resource("s3")
                bucket = s3.Bucket(s3_bucket)
                bucket.upload_file(f.name, s3_object_key, ExtraArgs=extra_args)
            except:
                s3_client.upload_file(f.name, s3_bucket, s3_object_key, ExtraArgs=extra_args)
            return s3_path_without_file

    def _create_athena_table_from_s3_path(self, s3_path: str, table_name: str, hive_columns: dict) -> str:
        athena_database = self._get_athena_database()
        query = CREATE_TABLE_TEMPLATE.render(
            database=athena_database,
            table=table_name,
            s3_location=s3_path,
            columns=hive_columns,
            offline_store_type="parquet",
        )
        logger.info(f"Creating Athena table {athena_database}.{table_name}...")

        self.sql(query)

        logger.info(f"Table {athena_database}.{table_name} was successfully created")

        # The database/table names are in the quotation marks to handle the scenario when they start with digits.
        return f'"{athena_database}"."{table_name}"'

    def _pandas_columns_to_hive_columns(self, pandas_df: pandas.DataFrame):
        column_types = {}
        for k, v in pandas_df.dtypes.to_dict().items():
            if "datetime64" in v.name:
                column_types[k] = "timestamp"
                continue
            type_name = v.name.lower()
            if type_name not in PANDAS_TO_HIVE_TYPES:
                raise Exception(f"Pandas Type {type_name} not supported. Mapping to Hive Type not found.")
            column_types[k] = PANDAS_TO_HIVE_TYPES[type_name]
        return column_types

    def write_pandas(self, df: pandas.DataFrame, table_name: str) -> str:
        s3_full_path = self._upload_pandas_to_s3(df)
        hive_columns = self._pandas_columns_to_hive_columns(df)

        return self._create_athena_table_from_s3_path(s3_full_path, table_name, hive_columns)

    def sql(self, sql_query: str) -> Union[str, Dict[str, Any]]:
        return self._wr.athena.start_query_execution(
            sql_query,
            s3_output=self._get_athena_s3_path(),
            database=self._get_athena_database(),
            wait=True,
            boto3_session=self.config.boto3_session,
            workgroup=self.config.workgroup,
            encryption=self.config.encryption,
            kms_key=self.config.kms_key,
        )

    def read_sql(self, sql_query: str) -> Union[pandas.DataFrame, Iterator[pandas.DataFrame]]:
        return self._wr.athena.read_sql_query(
            sql_query,
            s3_output=self._get_athena_s3_path(),
            database=self._get_athena_database(),
            boto3_session=self.config.boto3_session,
            workgroup=self.config.workgroup,
            encryption=self.config.encryption,
            kms_key=self.config.kms_key,
        )

    def get_database(self) -> str:
        return self._get_athena_database()

    def create_table(self, table_spec: AthenaTableCreationSpec):
        logger.info(f"Registering Glue table {table_spec.database}.{table_spec.table}...")
        sql = CREATE_TABLE_TEMPLATE.render(
            database=table_spec.database,
            table=table_spec.table,
            s3_location=table_spec.s3_location,
            columns=table_spec.columns,
            partition_by=table_spec.partition_by,
            partition_by_type=table_spec.partition_by_type,
            partition_by_format=table_spec.partition_by_format,
            partition_by_range_from=table_spec.partition_by_range_from,
            partition_by_range_to=table_spec.partition_by_range_to,
            partition_by_interval=table_spec.partition_by_interval,
            tecton_table_metadata_hash=table_spec.stable_hash,
            tecton_table_spec_version=AthenaTableCreationSpec.SPEC_VERSION,
            offline_store_type=table_spec.data_format,
        )
        self.sql(sql)

    def create_view(self, view_sql, view_name):
        logger.info(f"Creating view {view_name}.")
        sql = f"CREATE VIEW {view_name} as {view_sql}"
        self.sql(sql)

    def _is_existing_table_equivalent_to_spec(self, table_spec: AthenaTableCreationSpec):
        parameters = self._get_table_parameters(table_spec.database, table_spec.table)

        return parameters.get(GLUE_CATALOG_TABLE_PROPERTY_HASH) == table_spec.stable_hash

    def _get_existing_table_spec_version(self, table_spec: AthenaTableCreationSpec):
        parameters = self._get_table_parameters(table_spec.database, table_spec.table)
        if GLUE_CATALOG_TABLE_PROPERTY_SPEC_VERSION not in parameters:
            return None

        return int(parameters[GLUE_CATALOG_TABLE_PROPERTY_SPEC_VERSION])

    def create_table_if_necessary(self, table_spec: AthenaTableCreationSpec):
        if not self._does_table_exist(table_spec.database, table_spec.table):
            return self.create_table(table_spec)

        existing_table_spec_version = self._get_existing_table_spec_version(table_spec)

        if self._is_existing_table_equivalent_to_spec(table_spec):
            logger.debug(
                f"Glue table {table_spec.database}.{table_spec.table} hash matches expectation ({table_spec.stable_hash}). No update needed."
            )
            if table_spec.data_format == "delta":
                # required for partition discovery
                self.repair_table(table_spec.database, table_spec.table)
            return
        if existing_table_spec_version is None:
            raise TectonAthenaValidationError(
                f"Glue table {table_spec.database}.{table_spec.table} registration doesn't meet expectations but cannot be updated because it doesn't seem to have been registered by Tecton. Please drop the table to have Tecton manage its registration."
            )
        if existing_table_spec_version > AthenaTableCreationSpec.SPEC_VERSION:
            raise TectonAthenaValidationError(
                f"Glue table {table_spec.database}.{table_spec.table} registration doesn't meet expectations but cannot be updated because it was registered with a newer version of Tecton. Please upgrade the SDK. Found Spec version: {existing_table_spec_version}. SDK Spec version: {AthenaTableCreationSpec.SPEC_VERSION}"
            )
        logger.info(
            f"Glue table {table_spec.database}.{table_spec.table} registration needs to be updated. Dropping existing table..."
        )
        self.delete_table_if_exists(table_spec.database, table_spec.table)
        table_create_result = self.create_table(table_spec)
        return table_create_result

    def repair_table(self, database, table):
        self.sql(f"MSCK REPAIR TABLE `{database}`.`{table}`")

    def get_spine_temp_table_name(self):
        return "TMP_SPINE_TABLE_" + secrets.token_hex(8)

    def _get_athena_database(self):
        return self.config.database or conf.get_or_none("ATHENA_DATABASE") or "tecton_temp_tables"


_session = None


def get_session():
    global _session

    if _session is not None:
        return _session

    _session = AthenaSession()
    return _session



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/schema_container.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.common import schema_pb2 as tecton__proto_dot_common_dot_schema__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n*tecton_proto/common/schema_container.proto\x12\x13tecton_proto.common\x1a tecton_proto/common/schema.proto\"g\n\x0fSchemaContainer\x12\x42\n\rtecton_schema\x18\x01 \x01(\x0b\x32\x1b.tecton_proto.common.SchemaH\x00R\x0ctectonSchemaB\x10\n\x0eschema_optionsB\x15\n\x11\x63om.tecton.commonP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.common.schema_container_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\021com.tecton.commonP\001'
  _SCHEMACONTAINER._serialized_start=101
  _SCHEMACONTAINER._serialized_end=204
# @@protoc_insertion_point(module_scope)



from typing import Dict
from typing import List
from typing import Optional
from typing import Type

import attrs
import pendulum
from pyspark.sql import SparkSession

from tecton._internals import time_utils
from tecton_core import conf
from tecton_core.pipeline_common import get_time_window_from_data_source_node
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.node_interface import QueryNode
from tecton_core.query.nodes import AddAnchorTimeNode
from tecton_core.query.nodes import AddDurationNode
from tecton_core.query.nodes import AddEffectiveTimestampNode
from tecton_core.query.nodes import AsofJoinFullAggNode
from tecton_core.query.nodes import AsofJoinNode
from tecton_core.query.nodes import DataSourceScanNode
from tecton_core.query.nodes import EntityFilterNode
from tecton_core.query.nodes import FeatureTimeFilterNode
from tecton_core.query.nodes import FeatureViewPipelineNode
from tecton_core.query.nodes import MockDataSourceScanNode
from tecton_core.query.nodes import OfflineStoreScanNode
from tecton_core.query.nodes import PartialAggNode
from tecton_core.query.nodes import RenameColsNode
from tecton_core.query.nodes import RespectFeatureStartTimeNode
from tecton_core.query.nodes import SelectDistinctNode
from tecton_core.query.nodes import UserSpecifiedDataNode
from tecton_core.query.rewrite import Rewrite


def tree_contains(tree: NodeRef, node_type: Type[QueryNode]) -> bool:
    """Returns True if the tree contains a NodeRef of the given type, False otherwise."""
    if isinstance(tree.node, node_type):
        return True

    return any(tree_contains(subtree, node_type) for subtree in tree.inputs)


def find_node_of_type(tree: NodeRef, node_type: Type[QueryNode]) -> Optional[NodeRef]:
    """Returns the NodeRef (i.e. a subtree) of the given type, or None if no such NodeRef exists.

    Raises an AssertionError if there are multiple NodeRefs of the given type.
    """
    nodes = _find_all_nodes_of_type(tree, node_type)
    assert len(nodes) < 2, f"Expected at most one node of type {node_type}, found {len(nodes)}"
    return nodes[0] if len(nodes) == 1 else None


def _find_all_nodes_of_type(tree: NodeRef, node_type: Type[QueryNode]) -> List[NodeRef]:
    """Returns a list of all NodeRefs of the given type."""
    nodes = []

    if isinstance(tree.node, node_type):
        nodes.append(tree)

    for subtree in tree.inputs:
        nodes.extend(_find_all_nodes_of_type(subtree, node_type))

    return nodes


class MockDataRewrite(Rewrite):
    """
    Replace DataSourceScanNode with MockDataSourceScanNode based on DataSource ID map.
    """

    def __init__(self, mock_data: Dict[str, NodeRef]):
        self.mock_data = mock_data

    def rewrite(self, tree: NodeRef):
        if isinstance(tree.node, DataSourceScanNode):
            node = tree.node
            if node.ds.id in self.mock_data:
                # Replace with mock data
                tree.node = MockDataSourceScanNode(
                    self.mock_data[node.ds.id],
                    node.ds,
                    self.mock_data[node.ds.id].columns,
                    node.start_time,
                    node.end_time,
                )
        else:
            for i in tree.inputs:
                self.rewrite(i)


class SpineEntityPushdown(Rewrite):
    """Filters the original feature data with respect to the entities contained in a spine.

    Requires a Spark session since the spine must be evaluated.

    This should be applied to AsofJoinNodes and AsofJoinFullAggNodes since both of have sorts (as part of windows), for
    which we would like to minimize memory usage.
    """

    def __init__(self, spark: SparkSession):
        self.spark = spark

    def rewrite(self, tree: NodeRef):
        if isinstance(tree.node, AsofJoinNode):
            self.rewrite_asof(tree)
        elif isinstance(tree.node, AsofJoinFullAggNode):
            self.rewrite_asof_full_agg(tree)
        else:
            for i in tree.inputs:
                self.rewrite(i)

    def rewrite_asof(self, tree: NodeRef):
        node = tree.node
        # In our current usage of AsofJoinNode, we always pass in the spine on the left.
        spine_df = NodeRef.shallow_copy(node.left_container.node)
        self.pushdown_entities(node.right_container.node, spine_df, node.join_cols)

    def rewrite_asof_full_agg(self, tree: NodeRef):
        node = tree.node

        # We only want to rewrite for a user specified spine.
        if not tree_contains(node.spine, UserSpecifiedDataNode):
            return
        spine = NodeRef.shallow_copy(node.spine)
        self.pushdown_entities(node.partial_agg_node, spine, node.fdw.join_keys)

    def pushdown_entities(self, tree: NodeRef, spine: NodeRef, join_cols: List[str]):
        node = tree.node
        can_be_pushed_down = (
            RespectFeatureStartTimeNode,
            RenameColsNode,
            FeatureTimeFilterNode,
            AddAnchorTimeNode,
            AddDurationNode,
            AddEffectiveTimestampNode,
            PartialAggNode,
        )
        if isinstance(node, can_be_pushed_down):
            self.pushdown_entities(node.input_node, spine, join_cols)
        else:
            entities_node = SelectDistinctNode(spine, join_cols).as_ref()
            tree.node = EntityFilterNode(node.as_ref(), entities_node, join_cols)


# TODO: genericize this so it can be applied to non-spark. Right now we depend on directly being able to read dataframe from spark to get time limits.
class SpineTimePushdown(Rewrite):
    """
    We actually evaluate the spine to get the time ranges for this rewrite, so we need a spark session to do so.
    """

    def __init__(self, spark: Optional[SparkSession]):
        self.spark = spark
        # The spine time limits are the same for all spines used throughout the query, so we only calculate once.
        self.spine_time_limits: Optional[pendulum.Period] = None

    def rewrite(self, tree: NodeRef):
        if isinstance(tree.node, AsofJoinNode):
            self.rewrite_asof(tree)
        elif isinstance(tree.node, AsofJoinFullAggNode):
            self.rewrite_asof_full_agg(tree)
        else:
            for i in tree.inputs:
                self.rewrite(i)

    def rewrite_asof(self, tree: NodeRef):
        node = tree.node
        # In our current usage of the code, we always pass in the spine on the left side of the asof join.
        # This rewrite is still applicable for any type of dataframe on the left, but referring to it as spine
        # to make the naming match up with the aggregate case (in which the spine is not on the left).
        if self.spine_time_limits is None:
            cur_node = node.left_container.node.node
            self.spine_time_limits = _get_spine_time_limits(cur_node)

        self.pushdown_time_range(node.right_container.node, self.spine_time_limits)

    # Compute the time limits from the node, and push down the time limits to its input
    def rewrite_asof_full_agg(self, tree: NodeRef):
        node = tree.node

        # We only want to rewrite for a user specified spine.
        spine_node = find_node_of_type(node.spine, UserSpecifiedDataNode)
        if spine_node is None:
            return
        if self.spine_time_limits is None:
            self.spine_time_limits = _get_spine_time_limits(node)
        self.pushdown_time_range(node.partial_agg_node, self.spine_time_limits)

    # Push down and convert spine time filter to either raw data or feature time filter at the DataSourceScanNode or OfflineStoreScanNode.
    # Nodes that do not affect the correlation with the spine time range are enumerated in the can_be_pushed_down list.
    def pushdown_time_range(self, tree: NodeRef, spine_time_limits: pendulum.Period):
        node = tree.node
        can_be_pushed_down = (
            RespectFeatureStartTimeNode,
            RenameColsNode,
            PartialAggNode,
            FeatureTimeFilterNode,
            AddAnchorTimeNode,
            AddDurationNode,
            AddEffectiveTimestampNode,
        )
        if isinstance(node, can_be_pushed_down):
            self.pushdown_time_range(node.input_node, spine_time_limits)
        elif isinstance(node, (OfflineStoreScanNode, FeatureViewPipelineNode)):
            feature_time_limits = time_utils.get_feature_data_time_limits(
                fd=node.feature_definition_wrapper, spine_time_limits=spine_time_limits
            )
            if isinstance(node, FeatureViewPipelineNode):
                # swap the pipeline node to add new time limits
                node = FeatureViewPipelineNode(node.inputs_map, node.feature_definition_wrapper, feature_time_limits)
                tree.node = node

                for n in node.inputs:
                    if isinstance(n.node, DataSourceScanNode):
                        # this method will convert aligned_feature_time_limits to raw data time limits by accounting for FilteredSource offsets etc.
                        data_time_filter = get_time_window_from_data_source_node(
                            feature_time_limits,
                            node.feature_definition_wrapper.batch_materialization_schedule,
                            n.node.ds_node,
                        )
                        if data_time_filter is not None:
                            n.node = attrs.evolve(
                                n.node, start_time=data_time_filter.start, end_time=data_time_filter.end
                            )
            elif isinstance(node, OfflineStoreScanNode):
                tree.node = attrs.evolve(node, time_filter=feature_time_limits)


# Mutates the input
def rewrite_tree_for_spine(tree: NodeRef, spark: Optional[SparkSession]):
    if not conf.get_bool("QUERY_REWRITE_ENABLED"):
        return
    rewrites = [SpineTimePushdown(spark), SpineEntityPushdown(spark)]
    for rewrite in rewrites:
        rewrite.rewrite(tree)


def _get_spine_user_specified_data_node(cur_node: QueryNode) -> Optional[UserSpecifiedDataNode]:
    if isinstance(cur_node, UserSpecifiedDataNode):
        return cur_node

    # UserSpecifiedDataNode can be nested inside a spine subtree.
    # eg. The spine can be a SelectDistinctNode -> UserSpecifiedNode
    for child in cur_node.inputs:
        node = _get_spine_user_specified_data_node(child.node)
        if node:
            return node
    return None


def _get_spine_time_limits(cur_node: QueryNode) -> Optional[pendulum.Period]:
    user_specified_data_node = _get_spine_user_specified_data_node(cur_node)
    if not user_specified_data_node:
        return None
    timestamp_key = user_specified_data_node.metadata["timestamp_key"]
    return user_specified_data_node.data.get_time_range(timestamp_key)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/feature_view.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.args import pipeline_pb2 as tecton__proto_dot_args_dot_pipeline__pb2
from tecton_proto.args import feature_view_pb2 as tecton__proto_dot_args_dot_feature__view__pb2
from tecton_proto.common import aggregation_function_pb2 as tecton__proto_dot_common_dot_aggregation__function__pb2
from tecton_proto.data import fv_materialization_pb2 as tecton__proto_dot_data_dot_fv__materialization__pb2
from tecton_proto.common import schema_pb2 as tecton__proto_dot_common_dot_schema__pb2
from tecton_proto.common import data_source_type_pb2 as tecton__proto_dot_common_dot_data__source__type__pb2
from tecton_proto.common import framework_version_pb2 as tecton__proto_dot_common_dot_framework__version__pb2
from tecton_proto.data import fco_metadata_pb2 as tecton__proto_dot_data_dot_fco__metadata__pb2
from tecton_proto.common import id_pb2 as tecton__proto_dot_common_dot_id__pb2
from tecton_proto.validation import validator_pb2 as tecton__proto_dot_validation_dot_validator__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$tecton_proto/data/feature_view.proto\x12\x11tecton_proto.data\x1a\x1egoogle/protobuf/duration.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a tecton_proto/args/pipeline.proto\x1a$tecton_proto/args/feature_view.proto\x1a.tecton_proto/common/aggregation_function.proto\x1a*tecton_proto/data/fv_materialization.proto\x1a tecton_proto/common/schema.proto\x1a*tecton_proto/common/data_source_type.proto\x1a+tecton_proto/common/framework_version.proto\x1a$tecton_proto/data/fco_metadata.proto\x1a\x1ctecton_proto/common/id.proto\x1a\'tecton_proto/validation/validator.proto\"\xb1\r\n\x0b\x46\x65\x61tureView\x12?\n\x0f\x66\x65\x61ture_view_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\rfeatureViewId\x12\x41\n\x0c\x66\x63o_metadata\x18\x02 \x01(\x0b\x32\x1e.tecton_proto.data.FcoMetadataR\x0b\x66\x63oMetadata\x12\x36\n\nentity_ids\x18\x03 \x03(\x0b\x32\x17.tecton_proto.common.IdR\tentityIds\x12\x1b\n\tjoin_keys\x18\x04 \x03(\tR\x08joinKeys\x12?\n\x07schemas\x18\x0b \x01(\x0b\x32%.tecton_proto.data.FeatureViewSchemasR\x07schemas\x12P\n\x0b\x65nrichments\x18\xe8\x07 \x01(\x0b\x32).tecton_proto.data.FeatureViewEnrichmentsB\x02\x18\x01R\x0b\x65nrichments\x12U\n\x12temporal_aggregate\x18\x07 \x01(\x0b\x32$.tecton_proto.data.TemporalAggregateH\x00R\x11temporalAggregate\x12\x39\n\x08temporal\x18\x08 \x01(\x0b\x32\x1b.tecton_proto.data.TemporalH\x00R\x08temporal\x12]\n\x16on_demand_feature_view\x18\x11 \x01(\x0b\x32&.tecton_proto.data.OnDemandFeatureViewH\x00R\x13onDemandFeatureView\x12\x46\n\rfeature_table\x18\x10 \x01(\x0b\x32\x1f.tecton_proto.data.FeatureTableH\x00R\x0c\x66\x65\x61tureTable\x12#\n\rtimestamp_key\x18\n \x01(\tR\x0ctimestampKey\x12W\n\x14online_serving_index\x18\x05 \x01(\x0b\x32%.tecton_proto.data.OnlineServingIndexR\x12onlineServingIndex\x12\x37\n\x08pipeline\x18\x06 \x01(\x0b\x32\x1b.tecton_proto.args.PipelineR\x08pipeline\x12\x62\n\x16materialization_params\x18\t \x01(\x0b\x32+.tecton_proto.data.NewMaterializationParamsR\x15materializationParams\x12\x37\n\x17materialization_enabled\x18\x0c \x01(\x08R\x16materializationEnabled\x12}\n!materialization_state_transitions\x18\r \x03(\x0b\x32\x31.tecton_proto.data.MaterializationStateTransitionR\x1fmaterializationStateTransitions\x12P\n\x11monitoring_params\x18\x0e \x01(\x0b\x32#.tecton_proto.data.MonitoringParamsR\x10monitoringParams\x12?\n\x1c\x66\x65\x61ture_store_format_version\x18\x17 \x01(\x05R\x19\x66\x65\x61tureStoreFormatVersion\x12G\n\x0esnowflake_data\x18\x18 \x01(\x0b\x32 .tecton_proto.data.SnowflakeDataR\rsnowflakeData\x12/\n\x11\x66ramework_version\x18\x13 \x01(\x05\x42\x02\x18\x01R\x10\x66rameworkVersion\x12H\n\nfw_version\x18\x1a \x01(\x0e\x32%.tecton_proto.common.FrameworkVersionB\x02\x18\x01R\tfwVersion\x12\x17\n\x07web_url\x18\x19 \x01(\tR\x06webUrl\x12H\n\rbatch_trigger\x18\x1b \x01(\x0e\x32#.tecton_proto.args.BatchTriggerTypeR\x0c\x62\x61tchTrigger\x12[\n\x0fvalidation_args\x18\x1c \x01(\x0b\x32\x32.tecton_proto.validation.FeatureViewValidationArgsR\x0evalidationArgsB\x13\n\x11\x66\x65\x61ture_view_type\"\xbe\x02\n\x11TemporalAggregate\x12@\n\x0eslide_interval\x18\x01 \x01(\x0b\x32\x19.google.protobuf.DurationR\rslideInterval\x12\x32\n\x15slide_interval_string\x18\x05 \x01(\tR\x13slideIntervalString\x12?\n\x08\x66\x65\x61tures\x18\x02 \x03(\x0b\x32#.tecton_proto.data.AggregateFeatureR\x08\x66\x65\x61tures\x12#\n\ris_continuous\x18\x03 \x01(\x08R\x0cisContinuous\x12M\n\x10\x64\x61ta_source_type\x18\x04 \x01(\x0e\x32#.tecton_proto.common.DataSourceTypeR\x0e\x64\x61taSourceType\"\xc2\x02\n\x10\x41ggregateFeature\x12,\n\x12input_feature_name\x18\x01 \x01(\tR\x10inputFeatureName\x12.\n\x13output_feature_name\x18\x02 \x01(\tR\x11outputFeatureName\x12\x44\n\x08\x66unction\x18\x03 \x01(\x0e\x32(.tecton_proto.common.AggregationFunctionR\x08\x66unction\x12W\n\x0f\x66unction_params\x18\x05 \x01(\x0b\x32..tecton_proto.common.AggregationFunctionParamsR\x0e\x66unctionParams\x12\x31\n\x06window\x18\x04 \x01(\x0b\x32\x19.google.protobuf.DurationR\x06window\"\xf5\x01\n\x1dTrailingTimeWindowAggregation\x12\x19\n\x08time_key\x18\x01 \x01(\tR\x07timeKey\x12S\n\x18\x61ggregation_slide_period\x18\x02 \x01(\x0b\x32\x19.google.protobuf.DurationR\x16\x61ggregationSlidePeriod\x12?\n\x08\x66\x65\x61tures\x18\x03 \x03(\x0b\x32#.tecton_proto.data.AggregateFeatureR\x08\x66\x65\x61tures\x12#\n\ris_continuous\x18\x04 \x01(\x08R\x0cisContinuous\"\xbb\x02\n\x08Temporal\x12:\n\x0bserving_ttl\x18\x10 \x01(\x0b\x32\x19.google.protobuf.DurationR\nservingTtl\x12M\n\x10\x64\x61ta_source_type\x18\x11 \x01(\x0e\x32#.tecton_proto.common.DataSourceTypeR\x0e\x64\x61taSourceType\x12J\n\x0f\x62\x61\x63kfill_config\x18\x12 \x01(\x0b\x32!.tecton_proto.args.BackfillConfigR\x0e\x62\x61\x63kfillConfig\x12\x33\n\x15incremental_backfills\x18\x13 \x01(\x08R\x14incrementalBackfills\x12#\n\ris_continuous\x18\x14 \x01(\x08R\x0cisContinuous\"\x9a\x01\n\x0c\x46\x65\x61tureTable\x12%\n\x0eonline_enabled\x18\x01 \x01(\x08R\ronlineEnabled\x12\'\n\x0foffline_enabled\x18\x02 \x01(\x08R\x0eofflineEnabled\x12:\n\x0bserving_ttl\x18\x03 \x01(\x0b\x32\x19.google.protobuf.DurationR\nservingTtl\"*\n\x13OnDemandFeatureView\x12\x13\n\x05no_op\x18\x01 \x01(\x08R\x04noOp\"\xa6\x01\n\x12\x46\x65\x61tureViewSchemas\x12<\n\x0bview_schema\x18\x01 \x01(\x0b\x32\x1b.tecton_proto.common.SchemaR\nviewSchema\x12R\n\x16materialization_schema\x18\x02 \x01(\x0b\x32\x1b.tecton_proto.common.SchemaR\x15materializationSchema\"1\n\x12OnlineServingIndex\x12\x1b\n\tjoin_keys\x18\x01 \x03(\tR\x08joinKeys\"\xc4\x02\n\x1eMaterializationStateTransition\x12\x38\n\ttimestamp\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\ttimestamp\x12%\n\x0eonline_enabled\x18\x02 \x01(\x08R\ronlineEnabled\x12\'\n\x0foffline_enabled\x18\x03 \x01(\x08R\x0eofflineEnabled\x12R\n\x17\x66\x65\x61ture_start_timestamp\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x15\x66\x65\x61tureStartTimestamp\x12\x44\n\x1ematerialization_serial_version\x18\x05 \x01(\x05R\x1cmaterializationSerialVersion\"\xaf\x01\n\x19ParquetOfflineStoreParams\x12I\n\x13time_partition_size\x18\x01 \x01(\x0b\x32\x19.google.protobuf.DurationR\x11timePartitionSize\x12G\n\x07version\x18\x02 \x01(\x0e\x32-.tecton_proto.data.ParquetOfflineStoreVersionR\x07version\"\xab\x01\n\x17\x44\x65ltaOfflineStoreParams\x12I\n\x13time_partition_size\x18\x01 \x01(\x0b\x32\x19.google.protobuf.DurationR\x11timePartitionSize\x12\x45\n\x07version\x18\x02 \x01(\x0e\x32+.tecton_proto.data.DeltaOfflineStoreVersionR\x07version\"\xac\x01\n\x12OfflineStoreParams\x12H\n\x07parquet\x18\x01 \x01(\x0b\x32,.tecton_proto.data.ParquetOfflineStoreParamsH\x00R\x07parquet\x12\x42\n\x05\x64\x65lta\x18\x02 \x01(\x0b\x32*.tecton_proto.data.DeltaOfflineStoreParamsH\x00R\x05\x64\x65ltaB\x08\n\x06params\"\x98\t\n\x18NewMaterializationParams\x12\x46\n\x11schedule_interval\x18\x01 \x01(\x0b\x32\x19.google.protobuf.DurationR\x10scheduleInterval\x12\x62\n\x1fmaterialization_start_timestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x1dmaterializationStartTimestamp\x12R\n\x17\x66\x65\x61ture_start_timestamp\x18\r \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x15\x66\x65\x61tureStartTimestamp\x12^\n\x1emax_batch_aggregation_interval\x18\x03 \x01(\x0b\x32\x19.google.protobuf.DurationR\x1bmaxBatchAggregationInterval\x12\x33\n\x16writes_to_online_store\x18\x04 \x01(\x08R\x13writesToOnlineStore\x12\x35\n\x17writes_to_offline_store\x18\x05 \x01(\x08R\x14writesToOfflineStore\x12^\n\x14offline_store_config\x18\x06 \x01(\x0b\x32,.tecton_proto.args.OfflineFeatureStoreConfigR\x12offlineStoreConfig\x12W\n\x14offline_store_params\x18\x0e \x01(\x0b\x32%.tecton_proto.data.OfflineStoreParamsR\x12offlineStoreParams\x12U\n\x15\x62\x61tch_materialization\x18\x07 \x01(\x0b\x32 .tecton_proto.args.ClusterConfigR\x14\x62\x61tchMaterialization\x12W\n\x16stream_materialization\x18\x08 \x01(\x0b\x32 .tecton_proto.args.ClusterConfigR\x15streamMaterialization\x12L\n\x15max_source_data_delay\x18\t \x01(\x0b\x32\x19.google.protobuf.DurationR\x12maxSourceDataDelay\x12T\n\x13online_store_params\x18\n \x01(\x0b\x32$.tecton_proto.data.OnlineStoreParamsR\x11onlineStoreParams\x12\x44\n\routput_stream\x18\x0b \x01(\x0b\x32\x1f.tecton_proto.args.OutputStreamR\x0coutputStream\x12]\n\x11time_range_policy\x18\x0c \x01(\x0e\x32\x31.tecton_proto.data.MaterializationTimeRangePolicyR\x0ftimeRangePolicy\"\xc7\x01\n\x11OnlineStoreParams\x12%\n\x0euser_specified\x18\x01 \x01(\x08R\ruserSpecified\x12@\n\x06\x64ynamo\x18\x02 \x01(\x0b\x32&.tecton_proto.args.DynamoDbOnlineStoreH\x00R\x06\x64ynamo\x12;\n\x05redis\x18\x03 \x01(\x0b\x32#.tecton_proto.args.RedisOnlineStoreH\x00R\x05redisB\x0c\n\nstore_type\"\x92\x02\n\x10MonitoringParams\x12%\n\x0euser_specified\x18\x01 \x01(\x08R\ruserSpecified\x12+\n\x11monitor_freshness\x18\x02 \x01(\x08R\x10monitorFreshness\x12W\n\x1a\x65xpected_feature_freshness\x18\x03 \x01(\x0b\x32\x19.google.protobuf.DurationR\x18\x65xpectedFeatureFreshness\x12\x1f\n\x0b\x61lert_email\x18\x04 \x01(\tR\nalertEmail\x12\x30\n\x14grace_period_seconds\x18\x05 \x01(\x05R\x12gracePeriodSeconds\"\x85\x01\n\x16\x46\x65\x61tureViewEnrichments\x12S\n\x12\x66p_materialization\x18\x04 \x01(\x0b\x32$.tecton_proto.data.FvMaterializationR\x11\x66pMaterializationJ\x04\x08\x01\x10\x02J\x04\x08\x02\x10\x03J\x04\x08\x03\x10\x04J\x04\x08\x05\x10\x06\"?\n\rSnowflakeData\x12.\n\x13snowflake_view_name\x18\x01 \x01(\tR\x11snowflakeViewName*\x91\x01\n\x1aParquetOfflineStoreVersion\x12)\n%PARQUET_OFFLINE_STORE_VERSION_UNKNOWN\x10\x00\x12#\n\x1fPARQUET_OFFLINE_STORE_VERSION_1\x10\x01\x12#\n\x1fPARQUET_OFFLINE_STORE_VERSION_2\x10\x02*f\n\x18\x44\x65ltaOfflineStoreVersion\x12\'\n#DELTA_OFFLINE_STORE_VERSION_UNKNOWN\x10\x00\x12!\n\x1d\x44\x45LTA_OFFLINE_STORE_VERSION_1\x10\x01*\xc6\x01\n\x1eMaterializationTimeRangePolicy\x12\x31\n-MATERIALIZATION_TIME_RANGE_POLICY_UNSPECIFIED\x10\x00\x12:\n6MATERIALIZATION_TIME_RANGE_POLICY_FAIL_IF_OUT_OF_RANGE\x10\x01\x12\x35\n1MATERIALIZATION_TIME_RANGE_POLICY_FILTER_TO_RANGE\x10\x02\x42\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.feature_view_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _FEATUREVIEW.fields_by_name['enrichments']._options = None
  _FEATUREVIEW.fields_by_name['enrichments']._serialized_options = b'\030\001'
  _FEATUREVIEW.fields_by_name['framework_version']._options = None
  _FEATUREVIEW.fields_by_name['framework_version']._serialized_options = b'\030\001'
  _FEATUREVIEW.fields_by_name['fw_version']._options = None
  _FEATUREVIEW.fields_by_name['fw_version']._serialized_options = b'\030\001'
  _PARQUETOFFLINESTOREVERSION._serialized_start=6583
  _PARQUETOFFLINESTOREVERSION._serialized_end=6728
  _DELTAOFFLINESTOREVERSION._serialized_start=6730
  _DELTAOFFLINESTOREVERSION._serialized_end=6832
  _MATERIALIZATIONTIMERANGEPOLICY._serialized_start=6835
  _MATERIALIZATIONTIMERANGEPOLICY._serialized_end=7033
  _FEATUREVIEW._serialized_start=521
  _FEATUREVIEW._serialized_end=2234
  _TEMPORALAGGREGATE._serialized_start=2237
  _TEMPORALAGGREGATE._serialized_end=2555
  _AGGREGATEFEATURE._serialized_start=2558
  _AGGREGATEFEATURE._serialized_end=2880
  _TRAILINGTIMEWINDOWAGGREGATION._serialized_start=2883
  _TRAILINGTIMEWINDOWAGGREGATION._serialized_end=3128
  _TEMPORAL._serialized_start=3131
  _TEMPORAL._serialized_end=3446
  _FEATURETABLE._serialized_start=3449
  _FEATURETABLE._serialized_end=3603
  _ONDEMANDFEATUREVIEW._serialized_start=3605
  _ONDEMANDFEATUREVIEW._serialized_end=3647
  _FEATUREVIEWSCHEMAS._serialized_start=3650
  _FEATUREVIEWSCHEMAS._serialized_end=3816
  _ONLINESERVINGINDEX._serialized_start=3818
  _ONLINESERVINGINDEX._serialized_end=3867
  _MATERIALIZATIONSTATETRANSITION._serialized_start=3870
  _MATERIALIZATIONSTATETRANSITION._serialized_end=4194
  _PARQUETOFFLINESTOREPARAMS._serialized_start=4197
  _PARQUETOFFLINESTOREPARAMS._serialized_end=4372
  _DELTAOFFLINESTOREPARAMS._serialized_start=4375
  _DELTAOFFLINESTOREPARAMS._serialized_end=4546
  _OFFLINESTOREPARAMS._serialized_start=4549
  _OFFLINESTOREPARAMS._serialized_end=4721
  _NEWMATERIALIZATIONPARAMS._serialized_start=4724
  _NEWMATERIALIZATIONPARAMS._serialized_end=5900
  _ONLINESTOREPARAMS._serialized_start=5903
  _ONLINESTOREPARAMS._serialized_end=6102
  _MONITORINGPARAMS._serialized_start=6105
  _MONITORINGPARAMS._serialized_end=6379
  _FEATUREVIEWENRICHMENTS._serialized_start=6382
  _FEATUREVIEWENRICHMENTS._serialized_end=6515
  _SNOWFLAKEDATA._serialized_start=6517
  _SNOWFLAKEDATA._serialized_end=6580
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/databricks_api/execution.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n+tecton_proto/databricks_api/execution.proto\x12\x1btecton_proto.databricks_api\"Q\n\x15\x43ontextsCreateRequest\x12\x1a\n\x08language\x18\x01 \x01(\tR\x08language\x12\x1c\n\tclusterId\x18\x02 \x01(\tR\tclusterId\"(\n\x16\x43ontextsCreateResponse\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\"R\n\x14\x43ontextStatusRequest\x12\x1c\n\tclusterId\x18\x01 \x01(\tR\tclusterId\x12\x1c\n\tcontextId\x18\x02 \x01(\tR\tcontextId\"U\n\x15\x43ontextStatusResponse\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x16\n\x06status\x18\x02 \x01(\tR\x06status\x12\x14\n\x05\x65rror\x18\x03 \x01(\tR\x05\x65rror\"\x8a\x01\n\x16\x43ommandsExecuteRequest\x12\x1a\n\x08language\x18\x01 \x01(\tR\x08language\x12\x1c\n\tclusterId\x18\x02 \x01(\tR\tclusterId\x12\x1c\n\tcontextId\x18\x03 \x01(\tR\tcontextId\x12\x18\n\x07\x63ommand\x18\x04 \x01(\tR\x07\x63ommand\")\n\x17\x43ommandsExecuteResponse\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\"q\n\x15\x43ommandsStatusRequest\x12\x1c\n\tclusterId\x18\x01 \x01(\tR\tclusterId\x12\x1c\n\tcontextId\x18\x02 \x01(\tR\tcontextId\x12\x1c\n\tcommandId\x18\x03 \x01(\tR\tcommandId\"S\n\x07Results\x12\x12\n\x04\x64\x61ta\x18\x01 \x01(\tR\x04\x64\x61ta\x12\x1e\n\nresultType\x18\x02 \x01(\tR\nresultType\x12\x14\n\x05\x63\x61use\x18\x03 \x01(\tR\x05\x63\x61use\"p\n\x16\x43ommandsStatusResponse\x12>\n\x07results\x18\x01 \x01(\x0b\x32$.tecton_proto.databricks_api.ResultsR\x07results\x12\x16\n\x06status\x18\x02 \x01(\tR\x06status\"S\n\x15\x43ontextDestroyRequest\x12\x1c\n\tclusterId\x18\x01 \x01(\tR\tclusterId\x12\x1c\n\tcontextId\x18\x02 \x01(\tR\tcontextId\"(\n\x16\x43ontextDestroyResponse\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02idB\x1d\n\x19\x63om.tecton.databricks_apiP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.databricks_api.execution_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\031com.tecton.databricks_apiP\001'
  _CONTEXTSCREATEREQUEST._serialized_start=76
  _CONTEXTSCREATEREQUEST._serialized_end=157
  _CONTEXTSCREATERESPONSE._serialized_start=159
  _CONTEXTSCREATERESPONSE._serialized_end=199
  _CONTEXTSTATUSREQUEST._serialized_start=201
  _CONTEXTSTATUSREQUEST._serialized_end=283
  _CONTEXTSTATUSRESPONSE._serialized_start=285
  _CONTEXTSTATUSRESPONSE._serialized_end=370
  _COMMANDSEXECUTEREQUEST._serialized_start=373
  _COMMANDSEXECUTEREQUEST._serialized_end=511
  _COMMANDSEXECUTERESPONSE._serialized_start=513
  _COMMANDSEXECUTERESPONSE._serialized_end=554
  _COMMANDSSTATUSREQUEST._serialized_start=556
  _COMMANDSSTATUSREQUEST._serialized_end=669
  _RESULTS._serialized_start=671
  _RESULTS._serialized_end=754
  _COMMANDSSTATUSRESPONSE._serialized_start=756
  _COMMANDSSTATUSRESPONSE._serialized_end=868
  _CONTEXTDESTROYREQUEST._serialized_start=870
  _CONTEXTDESTROYREQUEST._serialized_end=953
  _CONTEXTDESTROYRESPONSE._serialized_start=955
  _CONTEXTDESTROYRESPONSE._serialized_end=995
# @@protoc_insertion_point(module_scope)



import datetime
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import pendulum
import pytimeparse
from google.protobuf import duration_pb2
from google.protobuf import timestamp_pb2
from typeguard import typechecked

from tecton_core.errors import TectonValidationError
from tecton_proto.data import feature_store_pb2


def timedelta_to_duration(td: datetime.timedelta) -> pendulum.Duration:
    return pendulum.duration(days=td.days, seconds=td.seconds, microseconds=td.microseconds)


def proto_to_duration(proto_duration: duration_pb2.Duration) -> pendulum.Duration:
    return timedelta_to_duration(proto_duration.ToTimedelta())


def timedelta_to_proto(td: Optional[datetime.timedelta]) -> Optional[duration_pb2.Duration]:
    if td is None:
        return None

    proto = duration_pb2.Duration()
    proto.FromTimedelta(td)
    return proto


def datetime_to_proto(dt: Optional[datetime.datetime]) -> Optional[timestamp_pb2.Timestamp]:
    if dt is None:
        return None

    proto = timestamp_pb2.Timestamp()
    proto.FromDatetime(dt)
    return proto


@typechecked
def to_human_readable_str(duration: Union[duration_pb2.Duration, datetime.timedelta]) -> str:
    """
    Replicate the backend logic that converts durations to feature strings, e.g. `column_sum_1h30m_30m`.

    Backend logic: https://github.com/tecton-ai/tecton/blob/e3c04e28066c21fa4ed9b459ece458c084766b6e/java/com/tecton/common/utils/TimeUtils.kt#L80
    """

    if isinstance(duration, duration_pb2.Duration):
        duration = proto_to_duration(duration)
    elif isinstance(duration, datetime.timedelta):
        duration = timedelta_to_duration(duration)
    else:
        raise TypeError(f"Invalid input duration type: {type(duration)}")

    duration_str_list = []
    if duration.days > 0:
        duration_str_list.append(f"{duration.days}d")
    if duration.hours > 0:
        duration_str_list.append(f"{duration.hours}h")
    if duration.minutes > 0:
        duration_str_list.append(f"{duration.minutes}m")
    if duration.remaining_seconds > 0:
        duration_str_list.append(f"{duration.remaining_seconds}s")

    return "0s" if len(duration_str_list) == 0 else "".join(duration_str_list)


def assert_is_round_seconds(d: pendulum.Duration) -> pendulum.Duration:
    if d % pendulum.Duration(seconds=1):
        raise ValueError(f"{d.in_words()} is not a round number of seconds")
    return d


def strict_pytimeparse(time_str: str) -> Union[int, float]:
    parsed = pytimeparse.parse(time_str)
    if parsed is None:
        raise TectonValidationError(f'Could not parse time string "{time_str}"')
    else:
        return parsed


def nanos_to_seconds(nanos: int) -> float:
    """
    :param nanos: Nanoseconds
    :return: Converts nanoseconds to seconds
    """
    return nanos / float(1e9)


def seconds_to_nanos(seconds: float) -> int:
    """
    :param seconds: Seconds
    :return: Converts seconds to nanoseconds
    """
    return int(seconds * 1e9)


def convert_duration_to_seconds(duration: int, version: int) -> int:
    """Convert a duration to seconds according to the version.

    If V0, the duration is already in seconds; if V1, the duration must be converted from nanos to seconds.
    """
    if version == feature_store_pb2.FEATURE_STORE_FORMAT_VERSION_DEFAULT:
        return duration
    else:
        return int(nanos_to_seconds(duration))


def convert_timedelta_for_version(duration: datetime.timedelta, version: int) -> int:
    """
    Convert pendulum duration according to version
    VO -> Return Seconds
    V1 -> Return Nanoseconds
    :param duration: Pendulum Duration
    :param version: Feature Store Format Version
    :return:
    """
    assert duration.microseconds == 0
    interval = duration.total_seconds()
    if version == feature_store_pb2.FEATURE_STORE_FORMAT_VERSION_DEFAULT:
        return int(interval)
    else:
        return seconds_to_nanos(interval)


def convert_epoch_to_datetime(epoch: int, version: int) -> pendulum.datetime:
    """
    Converts an epoch to a datetime.
    """
    if version == feature_store_pb2.FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_DEFAULT:
        return pendulum.from_timestamp(epoch)
    else:
        return pendulum.from_timestamp(epoch / 1e9)


def convert_timestamp_for_version(timestamp: datetime.datetime, version: int) -> int:
    if version == feature_store_pb2.FeatureStoreFormatVersion.FEATURE_STORE_FORMAT_VERSION_DEFAULT:
        return int(timestamp.timestamp())
    else:
        return seconds_to_nanos(timestamp.timestamp())


def convert_proto_duration_for_version(duration: duration_pb2.Duration, version: int) -> int:
    return convert_timedelta_for_version(duration.ToTimedelta(), version)


def align_time_downwards(time: datetime.datetime, alignment: datetime.timedelta) -> datetime.datetime:
    excess_seconds = time.timestamp() % alignment.total_seconds()
    return datetime.datetime.utcfromtimestamp(time.timestamp() - excess_seconds)


def align_epoch_downwards(int_timestamp_col: int, window_size: int) -> int:
    return int_timestamp_col - (int_timestamp_col % window_size)


def align_time_upwards(time: datetime.datetime, alignment: datetime.timedelta) -> datetime.datetime:
    excess_seconds = time.timestamp() % alignment.total_seconds()
    offset = 0 if excess_seconds == 0 else alignment.total_seconds() - excess_seconds
    return datetime.datetime.utcfromtimestamp(time.timestamp() + offset)


def get_timezone_aware_datetime(time: Optional[datetime.datetime]) -> Optional[datetime.datetime]:
    """
    Check and get timezone aware datetime
    :param time: Time to convert to timezone aware
    """
    if time is None:
        return None
    if time.tzinfo is None:
        return time.replace(tzinfo=pendulum.timezone("UTC"))
    else:
        return time


def backfill_jobs_periods(
    start_time: datetime.datetime, end_time: datetime.datetime, interval: datetime.timedelta
) -> List[Tuple[datetime.datetime, datetime.datetime]]:
    """
    Return a list of start/end tuples of size batch_schedule.
    For use of breaking up a large backfill window into incremental sizes.
    """
    jobs = []
    while start_time < end_time:
        jobs.append((start_time, start_time + interval))
        start_time = start_time + interval
    assert start_time == end_time, "Start and end times were not aligned to `interval`"
    return jobs



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Serializers for PyArrow and pandas conversions. See `pyspark.serializers` for more details.
"""

from pyspark.serializers import Serializer, read_int, write_int, UTF8Deserializer


class SpecialLengths(object):
    END_OF_DATA_SECTION = -1
    PYTHON_EXCEPTION_THROWN = -2
    TIMING_DATA = -3
    END_OF_STREAM = -4
    NULL = -5
    START_ARROW_STREAM = -6


class ArrowCollectSerializer(Serializer):
    """
    Deserialize a stream of batches followed by batch order information. Used in
    PandasConversionMixin._collect_as_arrow() after invoking Dataset.collectAsArrowToPython()
    in the JVM.
    """

    def __init__(self):
        self.serializer = ArrowStreamSerializer()

    def dump_stream(self, iterator, stream):
        return self.serializer.dump_stream(iterator, stream)

    def load_stream(self, stream):
        """
        Load a stream of un-ordered Arrow RecordBatches, where the last iteration yields
        a list of indices that can be used to put the RecordBatches in the correct order.
        """
        # load the batches
        for batch in self.serializer.load_stream(stream):
            yield batch

        # load the batch order indices or propagate any error that occurred in the JVM
        num = read_int(stream)
        if num == -1:
            error_msg = UTF8Deserializer().loads(stream)
            raise RuntimeError("An error occurred while calling "
                               "ArrowCollectSerializer.load_stream: {}".format(error_msg))
        batch_order = []
        for i in range(num):
            index = read_int(stream)
            batch_order.append(index)
        yield batch_order

    def __repr__(self):
        return "ArrowCollectSerializer(%s)" % self.serializer


class ArrowStreamSerializer(Serializer):
    """
    Serializes Arrow record batches as a stream.
    """

    def dump_stream(self, iterator, stream):
        import pyarrow as pa
        writer = None
        try:
            for batch in iterator:
                if writer is None:
                    writer = pa.RecordBatchStreamWriter(stream, batch.schema)
                writer.write_batch(batch)
        finally:
            if writer is not None:
                writer.close()

    def load_stream(self, stream):
        import pyarrow as pa
        reader = pa.ipc.open_stream(stream)
        for batch in reader:
            yield batch

    def __repr__(self):
        return "ArrowStreamSerializer"


class ArrowStreamPandasSerializer(ArrowStreamSerializer):
    """
    Serializes Pandas.Series as Arrow data with Arrow streaming format.

    Parameters
    ----------
    timezone : str
        A timezone to respect when handling timestamp values
    safecheck : bool
        If True, conversion from Arrow to Pandas checks for overflow/truncation
    assign_cols_by_name : bool
        If True, then Pandas DataFrames will get columns by name
    """

    def __init__(self, timezone, safecheck, assign_cols_by_name):
        super(ArrowStreamPandasSerializer, self).__init__()
        self._timezone = timezone
        self._safecheck = safecheck
        self._assign_cols_by_name = assign_cols_by_name

    def arrow_to_pandas(self, arrow_column):
        from pyspark.sql.pandas.types import _check_series_localize_timestamps, \
            _convert_map_items_to_dict
        import pyarrow

        # If the given column is a date type column, creates a series of datetime.date directly
        # instead of creating datetime64[ns] as intermediate data to avoid overflow caused by
        # datetime64[ns] type handling.
        s = arrow_column.to_pandas(date_as_object=True)

        if pyarrow.types.is_timestamp(arrow_column.type):
            return _check_series_localize_timestamps(s, self._timezone)
        elif pyarrow.types.is_map(arrow_column.type):
            return _convert_map_items_to_dict(s)
        else:
            return s

    def _create_batch(self, series):
        """
        Create an Arrow record batch from the given pandas.Series or list of Series,
        with optional type.

        Parameters
        ----------
        series : pandas.Series or list
            A single series, list of series, or list of (series, arrow_type)

        Returns
        -------
        pyarrow.RecordBatch
            Arrow RecordBatch
        """
        import pandas as pd
        import pyarrow as pa
        from pyspark.sql.pandas.types import _check_series_convert_timestamps_internal, \
            _convert_dict_to_map_items
        from pandas.api.types import is_categorical_dtype
        # Make input conform to [(series1, type1), (series2, type2), ...]
        if not isinstance(series, (list, tuple)) or \
                (len(series) == 2 and isinstance(series[1], pa.DataType)):
            series = [series]
        series = ((s, None) if not isinstance(s, (list, tuple)) else s for s in series)

        def create_array(s, t):
            mask = s.isnull()
            # Ensure timestamp series are in expected form for Spark internal representation
            if t is not None and pa.types.is_timestamp(t):
                s = _check_series_convert_timestamps_internal(s, self._timezone)
            elif t is not None and pa.types.is_map(t):
                s = _convert_dict_to_map_items(s)
            elif is_categorical_dtype(s.dtype):
                # Note: This can be removed once minimum pyarrow version is >= 0.16.1
                s = s.astype(s.dtypes.categories.dtype)
            try:
                array = pa.Array.from_pandas(s, mask=mask, type=t, safe=self._safecheck)
            except ValueError as e:
                if self._safecheck:
                    error_msg = "Exception thrown when converting pandas.Series (%s) to " + \
                                "Arrow Array (%s). It can be caused by overflows or other " + \
                                "unsafe conversions warned by Arrow. Arrow safe type check " + \
                                "can be disabled by using SQL config " + \
                                "`spark.sql.execution.pandas.convertToArrowArraySafely`."
                    raise ValueError(error_msg % (s.dtype, t)) from e
                else:
                    raise e
            return array

        arrs = []
        for s, t in series:
            if t is not None and pa.types.is_struct(t):
                if not isinstance(s, pd.DataFrame):
                    raise ValueError("A field of type StructType expects a pandas.DataFrame, "
                                     "but got: %s" % str(type(s)))

                # Input partition and result pandas.DataFrame empty, make empty Arrays with struct
                if len(s) == 0 and len(s.columns) == 0:
                    arrs_names = [(pa.array([], type=field.type), field.name) for field in t]
                # Assign result columns by schema name if user labeled with strings
                elif self._assign_cols_by_name and any(isinstance(name, str)
                                                       for name in s.columns):
                    arrs_names = [(create_array(s[field.name], field.type), field.name)
                                  for field in t]
                # Assign result columns by  position
                else:
                    arrs_names = [(create_array(s[s.columns[i]], field.type), field.name)
                                  for i, field in enumerate(t)]

                struct_arrs, struct_names = zip(*arrs_names)
                arrs.append(pa.StructArray.from_arrays(struct_arrs, struct_names))
            else:
                arrs.append(create_array(s, t))

        return pa.RecordBatch.from_arrays(arrs, ["_%d" % i for i in range(len(arrs))])

    def dump_stream(self, iterator, stream):
        """
        Make ArrowRecordBatches from Pandas Series and serialize. Input is a single series or
        a list of series accompanied by an optional pyarrow type to coerce the data to.
        """
        batches = (self._create_batch(series) for series in iterator)
        super(ArrowStreamPandasSerializer, self).dump_stream(batches, stream)

    def load_stream(self, stream):
        """
        Deserialize ArrowRecordBatches to an Arrow table and return as a list of pandas.Series.
        """
        batches = super(ArrowStreamPandasSerializer, self).load_stream(stream)
        import pyarrow as pa
        for batch in batches:
            yield [self.arrow_to_pandas(c) for c in pa.Table.from_batches([batch]).itercolumns()]

    def __repr__(self):
        return "ArrowStreamPandasSerializer"


class ArrowStreamPandasUDFSerializer(ArrowStreamPandasSerializer):
    """
    Serializer used by Python worker to evaluate Pandas UDFs
    """

    def __init__(self, timezone, safecheck, assign_cols_by_name, df_for_struct=False):
        super(ArrowStreamPandasUDFSerializer, self) \
            .__init__(timezone, safecheck, assign_cols_by_name)
        self._df_for_struct = df_for_struct

    def arrow_to_pandas(self, arrow_column):
        import pyarrow.types as types

        if self._df_for_struct and types.is_struct(arrow_column.type):
            import pandas as pd
            series = [super(ArrowStreamPandasUDFSerializer, self).arrow_to_pandas(column)
                      .rename(field.name)
                      for column, field in zip(arrow_column.flatten(), arrow_column.type)]
            s = pd.concat(series, axis=1)
        else:
            s = super(ArrowStreamPandasUDFSerializer, self).arrow_to_pandas(arrow_column)
        return s

    def dump_stream(self, iterator, stream):
        """
        Override because Pandas UDFs require a START_ARROW_STREAM before the Arrow stream is sent.
        This should be sent after creating the first record batch so in case of an error, it can
        be sent back to the JVM before the Arrow stream starts.
        """

        def init_stream_yield_batches():
            should_write_start_length = True
            for series in iterator:
                batch = self._create_batch(series)
                if should_write_start_length:
                    write_int(SpecialLengths.START_ARROW_STREAM, stream)
                    should_write_start_length = False
                yield batch

        return ArrowStreamSerializer.dump_stream(self, init_stream_yield_batches(), stream)

    def __repr__(self):
        return "ArrowStreamPandasUDFSerializer"


class CogroupUDFSerializer(ArrowStreamPandasUDFSerializer):

    def load_stream(self, stream):
        """
        Deserialize Cogrouped ArrowRecordBatches to a tuple of Arrow tables and yield as two
        lists of pandas.Series.
        """
        import pyarrow as pa
        dataframes_in_group = None

        while dataframes_in_group is None or dataframes_in_group > 0:
            dataframes_in_group = read_int(stream)

            if dataframes_in_group == 2:
                batch1 = [batch for batch in ArrowStreamSerializer.load_stream(self, stream)]
                batch2 = [batch for batch in ArrowStreamSerializer.load_stream(self, stream)]
                yield (
                    [self.arrow_to_pandas(c) for c in pa.Table.from_batches(batch1).itercolumns()],
                    [self.arrow_to_pandas(c) for c in pa.Table.from_batches(batch2).itercolumns()]
                )

            elif dataframes_in_group != 0:
                raise ValueError(
                    'Invalid number of pandas.DataFrames in group {0}'.format(dataframes_in_group))



from typing import Callable
from typing import List

import attr
from pyspark import SparkContext
from pyspark.sql import Column
from pyspark.sql import functions
from pyspark.sql import WindowSpec
from pyspark.sql.functions import expr

from tecton_core.aggregation_utils import get_aggregation_function_name
from tecton_core.aggregation_utils import sum_of_squares_column_prefix
from tecton_proto.common import aggregation_function_pb2 as afpb


# WARNING: If you're changing this class there's a good chance you need to change
# AggregationPlans.kt. Please look over that file carefully.


@attr.s(auto_attribs=True)
class AggregationPlan(object):
    """
    An AggregationPlan contains all the methods required to compute feature values for a specific Tecton aggregation.

    Partial aggregates are returned as a list of pyspark columns. Full aggregates use the partial aggregate columns as inputs and are returned as a single pyspark column.

    The order of the columns must be the same in:
    * the return list in partial_aggregation_transform
    * the arguments list in full_aggregation_transform
    * materialized_column_prefixes

    Attributes:
        partial_aggregation_transform: A method that maps an input column name to a list of output pyspark columns containing the partial aggregates.
        full_aggregation_transform: A method that maps a list of input partial aggregate columns and a WindowSpec to an output pyspark column containing the full aggregates.
        materialized_column_prefixes: The list of prefixes that should be applied to the pyspark columns produced by `partial_aggregation_transform`.
    """

    partial_aggregation_transform: Callable[[str], List[Column]]
    full_aggregation_transform: Callable[[List[str], WindowSpec], Column]
    materialized_column_prefixes: List[str]

    def materialized_column_names(self, input_column_name: str) -> List[str]:
        return [f"{prefix}_{input_column_name}" for prefix in self.materialized_column_prefixes]


def get_aggregation_plan(
    aggregation_function: afpb.AggregationFunction,
    function_params: afpb.AggregationFunctionParams,
    is_continuous: bool,
    time_key: str,
) -> AggregationPlan:
    plan = AGGREGATION_PLANS.get(aggregation_function, None)
    if plan is None:
        raise ValueError(f"Unsupported aggregation function {aggregation_function}")

    if callable(plan):
        return plan(time_key, function_params, is_continuous)
    else:
        return plan


def _simple_partial_aggregation_transform(spark_transform):
    return lambda col: [spark_transform(col)]


def _simple_full_aggregation_transform(spark_transform):
    return lambda cols, window: spark_transform(cols[0]).over(window)


def _simple_aggregation_plan(aggregation_function: afpb.AggregationFunction, spark_transform):
    return AggregationPlan(
        partial_aggregation_transform=_simple_partial_aggregation_transform(spark_transform),
        full_aggregation_transform=_simple_full_aggregation_transform(spark_transform),
        materialized_column_prefixes=[get_aggregation_function_name(aggregation_function)],
    )


# Partial aggregator used by first distinct N.
def FirstNAgg(timestamp: str, col: str, n: int) -> Column:
    sc = SparkContext._active_spark_context
    udf_name = f"tecton_first_distinct_{n}_partial_aggregation"
    sc._jvm.com.tecton.udfs.spark3.FirstNRegister().register(n, udf_name, True)
    return expr(f"{udf_name}({timestamp},{col}).values")


def _make_first_distinct_n_partial_aggregation(time_key: str, n: int) -> Callable:
    def _first_distinct_n_partial_aggregation(col: str) -> List[Column]:
        return [FirstNAgg(time_key, col, n)]

    return _first_distinct_n_partial_aggregation


# Partial aggregator used by last distinct N.
def LastDistinctNAgg(col1: str, col2: str, n: int) -> Column:
    sc = SparkContext._active_spark_context
    udf_name = f"tecton_last_distinct_{n}_partial_aggregation"
    sc._jvm.com.tecton.udfs.spark3.LastNRegister().register(n, udf_name, True)
    return expr(f"{udf_name}({col1}, {col2}).values")


def _make_last_distinct_n_partial_aggregation(time_key: str, n: int) -> Callable:
    def _last_distinct_n_partial_aggregation(col: str) -> List[Column]:
        return [LastDistinctNAgg(time_key, col, n)]

    return _last_distinct_n_partial_aggregation


# Full aggregator used by both last and first distinct N.
def LimitedListConcatAgg(col1: str, n: int, keep_last_items: bool) -> Column:
    sc = SparkContext._active_spark_context
    udf_name = (
        f"tecton_last_distinct_{n}_full_aggregation"
        if keep_last_items
        else f"tecton_first_distinct_{n}_full_aggregation"
    )
    sc._jvm.com.tecton.udfs.spark3.LimitedListConcatRegister().register(n, udf_name, keep_last_items)
    return expr(f"{udf_name}({col1})")


def _make_fixed_size_n_full_aggregation(n: int, keep_last_items: bool):
    def _fixed_size_n_full_aggregation(column_name: List[str], window: WindowSpec) -> Column:
        col = (LimitedListConcatAgg(column_name[0], n, keep_last_items).over(window)).values
        return col

    return _fixed_size_n_full_aggregation


# Full aggregator used by last non-distinct N.
def _make_last_non_distinct_n_full_aggregation(n: int) -> Callable:
    def _last_non_distinct_n_full_aggregation(cols: List[str], window: WindowSpec) -> Column:
        return functions.reverse(
            functions.slice(functions.reverse(functions.flatten(functions.collect_list(cols[0]).over(window))), 1, n)
        )

    return _last_non_distinct_n_full_aggregation


# Full aggregator used by first non-distinct N.
def _make_first_non_distinct_n_full_aggregation(n: int) -> Callable:
    def _first_non_distinct_n_full_aggregation(cols: List[Column], window: WindowSpec) -> Column:
        return functions.slice(functions.flatten(functions.collect_list(cols[0]).over(window)), 1, n)

    return _first_non_distinct_n_full_aggregation


# Partial aggregator used by last non-distinct N.
def _make_last_non_distinct_n_partial_aggregation(time_key: str, n: int) -> Callable:
    def last_non_distinct_n_partial_aggregation(col: str) -> List[Column]:
        # Sort items in descending order based on timestamp.
        sort_function = f"(left, right) -> case when left.{time_key} < right.{time_key} then 1 when left.{time_key} > right.{time_key} then -1 else 0 end)"
        return [
            functions.reverse(
                functions.slice(
                    functions.expr(f"array_sort(collect_list(struct({col}, {time_key})), {sort_function}"),
                    1,
                    n,
                ).getItem(col)
            )
        ]

    return last_non_distinct_n_partial_aggregation


# Partial aggregator used by first non-distinct N.
def _make_first_non_distinct_n_partial_aggregation(time_key: str, n: int) -> Callable:
    def first_non_distinct_n_partial_aggregation(col: str) -> List[Column]:
        # Sort items in ascending order based on timestamp.
        sort_function = f"(left,right) -> case when left.{time_key} < right.{time_key} then -1 when left.{time_key} > right.{time_key} then 1 else 0 end)"
        return [
            functions.slice(
                functions.expr(f"array_sort(collect_list(struct({col}, {time_key})), {sort_function}"),
                1,
                n,
            ).getItem(col)
        ]

    return first_non_distinct_n_partial_aggregation


def _sum_with_default(columns: List[str], window: WindowSpec):
    col = functions.sum(columns[0]).over(window)
    # Fill null
    col = functions.when(col.isNull(), functions.lit(0)).otherwise(col)
    return col


# population variation equation: Σ(x^2)/n - μ^2
def _var_pop_full_aggregation(cols: List[str], window: WindowSpec):
    sum_of_squares_col, count_col, sum_col = cols
    return functions.sum(sum_of_squares_col).over(window) / functions.sum(count_col).over(window) - functions.pow(
        functions.sum(sum_col).over(window) / functions.sum(count_col).over(window), 2
    )


def _stddev_pop_full_aggregation(cols: List[str], window: WindowSpec):
    return functions.sqrt(_var_pop_full_aggregation(cols, window))


# sample variation equation: (Σ(x^2) - (Σ(x)^2)/N)/N-1
def _var_samp_full_aggregation(cols: List[str], window: WindowSpec):
    sum_of_squares_col, count_col, sum_col = cols
    total_count_col = functions.sum(count_col).over(window)
    # check if count is equal to 0 for divide by 0 errors
    var_samp_col = functions.when(
        total_count_col != 1,
        (
            functions.sum(sum_of_squares_col).over(window)
            - functions.pow(functions.sum(sum_col).over(window), 2) / total_count_col
        )
        / (total_count_col - functions.lit(1)),
    )
    return var_samp_col


def _stddev_samp_full_aggregation(cols: List[str], window: WindowSpec):
    return functions.sqrt(_var_samp_full_aggregation(cols, window))


def _stddev_var_partial(col: List[str]):
    return [
        functions.sum(functions.pow(col, 2)),
        functions.count(col),
        functions.sum(functions.col(col)),
    ]


stddev_var_materialized_column_prefixes = [
    sum_of_squares_column_prefix,
    get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_COUNT),
    get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_SUM),
]

AGGREGATION_PLANS = {
    afpb.AGGREGATION_FUNCTION_SUM: _simple_aggregation_plan(afpb.AGGREGATION_FUNCTION_SUM, functions.sum),
    afpb.AGGREGATION_FUNCTION_MIN: _simple_aggregation_plan(afpb.AGGREGATION_FUNCTION_MIN, functions.min),
    afpb.AGGREGATION_FUNCTION_MAX: _simple_aggregation_plan(afpb.AGGREGATION_FUNCTION_MAX, functions.max),
    afpb.AGGREGATION_FUNCTION_LAST: _simple_aggregation_plan(
        afpb.AGGREGATION_FUNCTION_LAST, lambda col: functions.last(col, ignorenulls=True)
    ),
    # Needs to use COUNT for partial and SUM for full aggregation
    afpb.AGGREGATION_FUNCTION_COUNT: AggregationPlan(
        partial_aggregation_transform=_simple_partial_aggregation_transform(functions.count),
        full_aggregation_transform=_sum_with_default,
        materialized_column_prefixes=[get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_COUNT)],
    ),
    afpb.AGGREGATION_FUNCTION_LAST_DISTINCT_N: lambda time_key, params, is_continuous: AggregationPlan(
        partial_aggregation_transform=_make_last_distinct_n_partial_aggregation(time_key, params.last_n.n),
        full_aggregation_transform=_make_fixed_size_n_full_aggregation(params.last_n.n, True),
        materialized_column_prefixes=[
            get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_LAST_DISTINCT_N)
            + (str(params.last_n.n) if not is_continuous else "")
        ],
    ),
    afpb.AGGREGATION_FUNCTION_LAST_NON_DISTINCT_N: lambda time_key, params, is_continuous: AggregationPlan(
        partial_aggregation_transform=_make_last_non_distinct_n_partial_aggregation(time_key, params.last_n.n),
        full_aggregation_transform=_make_last_non_distinct_n_full_aggregation(params.last_n.n),
        materialized_column_prefixes=[
            get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_LAST_NON_DISTINCT_N)
            + (str(params.last_n.n) if not is_continuous else "")
        ],
    ),
    afpb.AGGREGATION_FUNCTION_VAR_POP: AggregationPlan(
        partial_aggregation_transform=lambda col: _stddev_var_partial(col),
        full_aggregation_transform=_var_pop_full_aggregation,
        materialized_column_prefixes=stddev_var_materialized_column_prefixes,
    ),
    afpb.AGGREGATION_FUNCTION_STDDEV_POP: AggregationPlan(
        partial_aggregation_transform=lambda col: _stddev_var_partial(col),
        full_aggregation_transform=_stddev_pop_full_aggregation,
        materialized_column_prefixes=stddev_var_materialized_column_prefixes,
    ),
    afpb.AGGREGATION_FUNCTION_VAR_SAMP: AggregationPlan(
        partial_aggregation_transform=lambda col: _stddev_var_partial(col),
        full_aggregation_transform=_var_samp_full_aggregation,
        materialized_column_prefixes=stddev_var_materialized_column_prefixes,
    ),
    afpb.AGGREGATION_FUNCTION_STDDEV_SAMP: AggregationPlan(
        partial_aggregation_transform=lambda col: _stddev_var_partial(col),
        full_aggregation_transform=_stddev_samp_full_aggregation,
        materialized_column_prefixes=stddev_var_materialized_column_prefixes,
    ),
    afpb.AGGREGATION_FUNCTION_FIRST_NON_DISTINCT_N: lambda time_key, params, is_continuous: AggregationPlan(
        partial_aggregation_transform=_make_first_non_distinct_n_partial_aggregation(time_key, params.first_n.n),
        full_aggregation_transform=_make_first_non_distinct_n_full_aggregation(params.first_n.n),
        materialized_column_prefixes=[
            get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_FIRST_NON_DISTINCT_N)
            + (str(params.first_n.n) if not is_continuous else "")
        ],
    ),
    afpb.AGGREGATION_FUNCTION_FIRST_DISTINCT_N: lambda time_key, params, is_continuous: AggregationPlan(
        partial_aggregation_transform=_make_first_distinct_n_partial_aggregation(time_key, params.first_n.n),
        full_aggregation_transform=_make_fixed_size_n_full_aggregation(params.first_n.n, False),
        materialized_column_prefixes=[
            get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_FIRST_DISTINCT_N)
            + (str(params.first_n.n) if not is_continuous else "")
        ],
    ),
}


def _mean_full_aggregation(cols: List[str], window: WindowSpec):
    # Window aggregation doesn't work with more than one built-in function like this
    #   sum(mean_clicked * count_clicked) / sum(count_clicked)
    # And it does not support UDFs on bounded windows (the kind we use)
    #   https://issues.apache.org/jira/browse/SPARK-22239
    # We work around this limitations by calculating ratio over two window aggregations
    mean_col, count_col = cols
    return functions.sum(functions.col(mean_col) * functions.col(count_col)).over(window) / functions.sum(
        count_col
    ).over(window)


# It is important that `partial_aggregation_transform` or `materialized_column_prefixes`
# contain aggregation data in the same ordering.
AGGREGATION_PLANS[afpb.AGGREGATION_FUNCTION_MEAN] = AggregationPlan(
    partial_aggregation_transform=lambda col: [functions.mean(col), functions.count(col)],
    full_aggregation_transform=_mean_full_aggregation,
    materialized_column_prefixes=[
        get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_MEAN),
        get_aggregation_function_name(afpb.AGGREGATION_FUNCTION_COUNT),
    ],
)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/dataobs/expectation.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.dataobs import metric_pb2 as tecton__proto_dot_dataobs_dot_metric__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n&tecton_proto/dataobs/expectation.proto\x12\x14tecton_proto.dataobs\x1a\x1fgoogle/protobuf/timestamp.proto\x1a!tecton_proto/dataobs/metric.proto\"\xb3\x02\n\x12\x46\x65\x61tureExpectation\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x1e\n\nexpression\x18\x02 \x01(\tR\nexpression\x12\x34\n\x16\x61lert_message_template\x18\x03 \x01(\tR\x14\x61lertMessageTemplate\x12,\n\x12input_column_names\x18\x04 \x03(\tR\x10inputColumnNames\x12?\n\rcreation_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0c\x63reationTime\x12\x44\n\x10last_update_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0elastUpdateTime\"\xce\x02\n\x11MetricExpectation\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x1e\n\nexpression\x18\x02 \x01(\tR\nexpression\x12\x34\n\x16\x61lert_message_template\x18\x03 \x01(\tR\x14\x61lertMessageTemplate\x12H\n\rinput_metrics\x18\x04 \x03(\x0b\x32#.tecton_proto.dataobs.FeatureMetricR\x0cinputMetrics\x12?\n\rcreation_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0c\x63reationTime\x12\x44\n\x10last_update_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0elastUpdateTimeBA\n\x12\x63om.tecton.dataobsP\x01Z)github.com/tecton-ai/tecton_proto/dataobs')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.dataobs.expectation_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\022com.tecton.dataobsP\001Z)github.com/tecton-ai/tecton_proto/dataobs'
  _FEATUREEXPECTATION._serialized_start=133
  _FEATUREEXPECTATION._serialized_end=440
  _METRICEXPECTATION._serialized_start=443
  _METRICEXPECTATION._serialized_end=777
# @@protoc_insertion_point(module_scope)



from enum import Enum

from tecton_core import conf


class ValidationMode(str, Enum):
    EXPLICIT = "explicit"
    AUTOMATIC = "auto"


def set_validation_mode(mode: ValidationMode):
    """Convenience utility to set the Tecton object validation mode for the lifetime of the Python process.

    Must be either "explicit" (tecton.ValidationMode.EXPLICIT) or "auto" (tecton.ValidationMode.AUTOMATIC). "explicit"
    is the default.

    In "auto" mode, locally defined Tecton objects will automatically trigger validation when needed.

    .. code-block:: python

        import tecton

        tecton.set_validation_mode("auto")

        credit_scores_batch = tecton.BatchSource(
            name='credit_scores_batch',
            batch_config=tecton.HiveConfig(database='demo_fraud', table='credit_scores'),
        )

        df = credit_scores_batch.get_dataframe()  # Will automatically trigger validation.

    In "explicit" mode, locally defined Tecton objects must be validated before they can be used to execute many
    methods.

    .. code-block:: python

        import tecton

        tecton.set_validation_mode("auto")

        credit_scores_batch = tecton.BatchSource(
            name='credit_scores_batch',
            batch_config=tecton.HiveConfig(database='demo_fraud', table='credit_scores'),
        )

        credit_scores_batch.validate()

        df = credit_scores_batch.get_dataframe()


    Note: Tecton objects fetched from the Tecton backend have already been validated during `tecton plan` and do
    not need to be re-validated.
    """
    if mode is None or mode.lower() not in (
        ValidationMode.AUTOMATIC,
        ValidationMode.EXPLICIT,
    ):
        raise ValueError(f"Mode should be one of 'auto' or 'explicit', got {mode}")
    conf.set("TECTON_VALIDATION_MODE", mode)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["StorageLevel"]


class StorageLevel(object):

    """
    Flags for controlling the storage of an RDD. Each StorageLevel records whether to use memory,
    whether to drop the RDD to disk if it falls out of memory, whether to keep the data in memory
    in a JAVA-specific serialized format, and whether to replicate the RDD partitions on multiple
    nodes. Also contains static constants for some commonly used storage levels, MEMORY_ONLY.
    Since the data is always serialized on the Python side, all the constants use the serialized
    formats.
    """

    def __init__(self, useDisk, useMemory, useOffHeap, deserialized, replication=1):
        self.useDisk = useDisk
        self.useMemory = useMemory
        self.useOffHeap = useOffHeap
        self.deserialized = deserialized
        self.replication = replication

    def __repr__(self):
        return "StorageLevel(%s, %s, %s, %s, %s)" % (
            self.useDisk, self.useMemory, self.useOffHeap, self.deserialized, self.replication)

    def __str__(self):
        result = ""
        result += "Disk " if self.useDisk else ""
        result += "Memory " if self.useMemory else ""
        result += "OffHeap " if self.useOffHeap else ""
        result += "Deserialized " if self.deserialized else "Serialized "
        result += "%sx Replicated" % self.replication
        return result

StorageLevel.DISK_ONLY = StorageLevel(True, False, False, False)
StorageLevel.DISK_ONLY_2 = StorageLevel(True, False, False, False, 2)
StorageLevel.DISK_ONLY_3 = StorageLevel(True, False, False, False, 3)
StorageLevel.MEMORY_ONLY = StorageLevel(False, True, False, False)
StorageLevel.MEMORY_ONLY_2 = StorageLevel(False, True, False, False, 2)
StorageLevel.MEMORY_AND_DISK = StorageLevel(True, True, False, False)
StorageLevel.MEMORY_AND_DISK_2 = StorageLevel(True, True, False, False, 2)
StorageLevel.OFF_HEAP = StorageLevel(True, True, True, False, 1)
StorageLevel.MEMORY_AND_DISK_DESER = StorageLevel(True, True, False, True)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import array
from abc import ABCMeta
import copy

import numpy as np
from py4j.java_gateway import JavaObject

from pyspark.ml.linalg import DenseVector, Vector, Matrix
from pyspark.ml.util import Identifiable


__all__ = ['Param', 'Params', 'TypeConverters']


class Param(object):
    """
    A param with self-contained documentation.

    .. versionadded:: 1.3.0
    """

    def __init__(self, parent, name, doc, typeConverter=None):
        if not isinstance(parent, Identifiable):
            raise TypeError("Parent must be an Identifiable but got type %s." % type(parent))
        self.parent = parent.uid
        self.name = str(name)
        self.doc = str(doc)
        self.typeConverter = TypeConverters.identity if typeConverter is None else typeConverter

    def _copy_new_parent(self, parent):
        """Copy the current param to a new parent, must be a dummy param."""
        if self.parent == "undefined":
            param = copy.copy(self)
            param.parent = parent.uid
            return param
        else:
            raise ValueError("Cannot copy from non-dummy parent %s." % parent)

    def __str__(self):
        return str(self.parent) + "__" + self.name

    def __repr__(self):
        return "Param(parent=%r, name=%r, doc=%r)" % (self.parent, self.name, self.doc)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if isinstance(other, Param):
            return self.parent == other.parent and self.name == other.name
        else:
            return False


class TypeConverters(object):
    """
    Factory methods for common type conversion functions for `Param.typeConverter`.

    .. versionadded:: 2.0.0
    """

    @staticmethod
    def _is_numeric(value):
        vtype = type(value)
        return vtype in [int, float, np.float64, np.int64] or vtype.__name__ == 'long'

    @staticmethod
    def _is_integer(value):
        return TypeConverters._is_numeric(value) and float(value).is_integer()

    @staticmethod
    def _can_convert_to_list(value):
        vtype = type(value)
        return vtype in [list, np.ndarray, tuple, range, array.array] or isinstance(value, Vector)

    @staticmethod
    def _can_convert_to_string(value):
        vtype = type(value)
        return isinstance(value, str) or vtype in [np.unicode_, np.string_, np.str_]

    @staticmethod
    def identity(value):
        """
        Dummy converter that just returns value.
        """
        return value

    @staticmethod
    def toList(value):
        """
        Convert a value to a list, if possible.
        """
        if type(value) == list:
            return value
        elif type(value) in [np.ndarray, tuple, range, array.array]:
            return list(value)
        elif isinstance(value, Vector):
            return list(value.toArray())
        else:
            raise TypeError("Could not convert %s to list" % value)

    @staticmethod
    def toListFloat(value):
        """
        Convert a value to list of floats, if possible.
        """
        if TypeConverters._can_convert_to_list(value):
            value = TypeConverters.toList(value)
            if all(map(lambda v: TypeConverters._is_numeric(v), value)):
                return [float(v) for v in value]
        raise TypeError("Could not convert %s to list of floats" % value)

    @staticmethod
    def toListListFloat(value):
        """
        Convert a value to list of list of floats, if possible.
        """
        if TypeConverters._can_convert_to_list(value):
            value = TypeConverters.toList(value)
            return [TypeConverters.toListFloat(v) for v in value]
        raise TypeError("Could not convert %s to list of list of floats" % value)

    @staticmethod
    def toListInt(value):
        """
        Convert a value to list of ints, if possible.
        """
        if TypeConverters._can_convert_to_list(value):
            value = TypeConverters.toList(value)
            if all(map(lambda v: TypeConverters._is_integer(v), value)):
                return [int(v) for v in value]
        raise TypeError("Could not convert %s to list of ints" % value)

    @staticmethod
    def toListString(value):
        """
        Convert a value to list of strings, if possible.
        """
        if TypeConverters._can_convert_to_list(value):
            value = TypeConverters.toList(value)
            if all(map(lambda v: TypeConverters._can_convert_to_string(v), value)):
                return [TypeConverters.toString(v) for v in value]
        raise TypeError("Could not convert %s to list of strings" % value)

    @staticmethod
    def toVector(value):
        """
        Convert a value to a MLlib Vector, if possible.
        """
        if isinstance(value, Vector):
            return value
        elif TypeConverters._can_convert_to_list(value):
            value = TypeConverters.toList(value)
            if all(map(lambda v: TypeConverters._is_numeric(v), value)):
                return DenseVector(value)
        raise TypeError("Could not convert %s to vector" % value)

    @staticmethod
    def toMatrix(value):
        """
        Convert a value to a MLlib Matrix, if possible.
        """
        if isinstance(value, Matrix):
            return value
        raise TypeError("Could not convert %s to matrix" % value)

    @staticmethod
    def toFloat(value):
        """
        Convert a value to a float, if possible.
        """
        if TypeConverters._is_numeric(value):
            return float(value)
        else:
            raise TypeError("Could not convert %s to float" % value)

    @staticmethod
    def toInt(value):
        """
        Convert a value to an int, if possible.
        """
        if TypeConverters._is_integer(value):
            return int(value)
        else:
            raise TypeError("Could not convert %s to int" % value)

    @staticmethod
    def toString(value):
        """
        Convert a value to a string, if possible.
        """
        if isinstance(value, str):
            return value
        elif type(value) in [np.string_, np.str_, np.unicode_]:
            return str(value)
        else:
            raise TypeError("Could not convert %s to string type" % type(value))

    @staticmethod
    def toBoolean(value):
        """
        Convert a value to a boolean, if possible.
        """
        if type(value) == bool:
            return value
        else:
            raise TypeError("Boolean Param requires value of type bool. Found %s." % type(value))


class Params(Identifiable, metaclass=ABCMeta):
    """
    Components that take parameters. This also provides an internal
    param map to store parameter values attached to the instance.

    .. versionadded:: 1.3.0
    """

    def __init__(self):
        super(Params, self).__init__()
        #: internal param map for user-supplied values param map
        self._paramMap = {}

        #: internal param map for default values
        self._defaultParamMap = {}

        #: value returned by :py:func:`params`
        self._params = None

        # Copy the params from the class to the object
        self._copy_params()

    def _copy_params(self):
        """
        Copy all params defined on the class to current object.
        """
        cls = type(self)
        src_name_attrs = [(x, getattr(cls, x)) for x in dir(cls)]
        src_params = list(filter(lambda nameAttr: isinstance(nameAttr[1], Param), src_name_attrs))
        for name, param in src_params:
            setattr(self, name, param._copy_new_parent(self))

    @property
    def params(self):
        """
        Returns all params ordered by name. The default implementation
        uses :py:func:`dir` to get all attributes of type
        :py:class:`Param`.
        """
        if self._params is None:
            self._params = list(filter(lambda attr: isinstance(attr, Param),
                                       [getattr(self, x) for x in dir(self) if x != "params" and
                                        not isinstance(getattr(type(self), x, None), property)]))
        return self._params

    def explainParam(self, param):
        """
        Explains a single param and returns its name, doc, and optional
        default value and user-supplied value in a string.
        """
        param = self._resolveParam(param)
        values = []
        if self.isDefined(param):
            if param in self._defaultParamMap:
                values.append("default: %s" % self._defaultParamMap[param])
            if param in self._paramMap:
                values.append("current: %s" % self._paramMap[param])
        else:
            values.append("undefined")
        valueStr = "(" + ", ".join(values) + ")"
        return "%s: %s %s" % (param.name, param.doc, valueStr)

    def explainParams(self):
        """
        Returns the documentation of all params with their optionally
        default values and user-supplied values.
        """
        return "\n".join([self.explainParam(param) for param in self.params])

    def getParam(self, paramName):
        """
        Gets a param by its name.
        """
        param = getattr(self, paramName)
        if isinstance(param, Param):
            return param
        else:
            raise ValueError("Cannot find param with name %s." % paramName)

    def isSet(self, param):
        """
        Checks whether a param is explicitly set by user.
        """
        param = self._resolveParam(param)
        return param in self._paramMap

    def hasDefault(self, param):
        """
        Checks whether a param has a default value.
        """
        param = self._resolveParam(param)
        return param in self._defaultParamMap

    def isDefined(self, param):
        """
        Checks whether a param is explicitly set by user or has
        a default value.
        """
        return self.isSet(param) or self.hasDefault(param)

    def hasParam(self, paramName):
        """
        Tests whether this instance contains a param with a given
        (string) name.
        """
        if isinstance(paramName, str):
            p = getattr(self, paramName, None)
            return isinstance(p, Param)
        else:
            raise TypeError("hasParam(): paramName must be a string")

    def getOrDefault(self, param):
        """
        Gets the value of a param in the user-supplied param map or its
        default value. Raises an error if neither is set.
        """
        param = self._resolveParam(param)
        if param in self._paramMap:
            return self._paramMap[param]
        else:
            return self._defaultParamMap[param]

    def extractParamMap(self, extra=None):
        """
        Extracts the embedded default param values and user-supplied
        values, and then merges them with extra values from input into
        a flat param map, where the latter value is used if there exist
        conflicts, i.e., with ordering: default param values <
        user-supplied values < extra.

        Parameters
        ----------
        extra : dict, optional
            extra param values

        Returns
        -------
        dict
            merged param map
        """
        if extra is None:
            extra = dict()
        paramMap = self._defaultParamMap.copy()
        paramMap.update(self._paramMap)
        paramMap.update(extra)
        return paramMap

    def copy(self, extra=None):
        """
        Creates a copy of this instance with the same uid and some
        extra params. The default implementation creates a
        shallow copy using :py:func:`copy.copy`, and then copies the
        embedded and extra parameters over and returns the copy.
        Subclasses should override this method if the default approach
        is not sufficient.

        Parameters
        ----------
        extra : dict, optional
            Extra parameters to copy to the new instance

        Returns
        -------
        :py:class:`Params`
            Copy of this instance
        """
        if extra is None:
            extra = dict()
        that = copy.copy(self)
        that._paramMap = {}
        that._defaultParamMap = {}
        return self._copyValues(that, extra)

    def set(self, param, value):
        """
        Sets a parameter in the embedded param map.
        """
        self._shouldOwn(param)
        try:
            value = param.typeConverter(value)
        except ValueError as e:
            raise ValueError('Invalid param value given for param "%s". %s' % (param.name, e))
        self._paramMap[param] = value

    def _shouldOwn(self, param):
        """
        Validates that the input param belongs to this Params instance.
        """
        if not (self.uid == param.parent and self.hasParam(param.name)):
            raise ValueError("Param %r does not belong to %r." % (param, self))

    def _resolveParam(self, param):
        """
        Resolves a param and validates the ownership.

        Parameters
        ----------
        param : str or :py:class:`Param`
            param name or the param instance, which must
            belong to this Params instance

        Returns
        -------
        :py:class:`Param`
            resolved param instance
        """
        if isinstance(param, Param):
            self._shouldOwn(param)
            return param
        elif isinstance(param, str):
            return self.getParam(param)
        else:
            raise ValueError("Cannot resolve %r as a param." % param)

    def _testOwnParam(self, param_parent, param_name):
        """
        Test the ownership. Return True or False
        """
        return self.uid == param_parent and self.hasParam(param_name)

    @staticmethod
    def _dummy():
        """
        Returns a dummy Params instance used as a placeholder to
        generate docs.
        """
        dummy = Params()
        dummy.uid = "undefined"
        return dummy

    def _set(self, **kwargs):
        """
        Sets user-supplied params.
        """
        for param, value in kwargs.items():
            p = getattr(self, param)
            if value is not None:
                try:
                    value = p.typeConverter(value)
                except TypeError as e:
                    raise TypeError('Invalid param value given for param "%s". %s' % (p.name, e))
            self._paramMap[p] = value
        return self

    def clear(self, param):
        """
        Clears a param from the param map if it has been explicitly set.
        """
        if self.isSet(param):
            del self._paramMap[param]

    def _setDefault(self, **kwargs):
        """
        Sets default params.
        """
        for param, value in kwargs.items():
            p = getattr(self, param)
            if value is not None and not isinstance(value, JavaObject):
                try:
                    value = p.typeConverter(value)
                except TypeError as e:
                    raise TypeError('Invalid default param value given for param "%s". %s'
                                    % (p.name, e))
            self._defaultParamMap[p] = value
        return self

    def _copyValues(self, to, extra=None):
        """
        Copies param values from this instance to another instance for
        params shared by them.

        Parameters
        ----------
        to : :py:class:`Params`
            the target instance
        extra : dict, optional
            extra params to be copied

        Returns
        -------
        :py:class:`Params`
            the target instance with param values copied
        """
        paramMap = self._paramMap.copy()
        if isinstance(extra, dict):
            for param, value in extra.items():
                if isinstance(param, Param):
                    paramMap[param] = value
                else:
                    raise TypeError("Expecting a valid instance of Param, but received: {}"
                                    .format(param))
        elif extra is not None:
            raise TypeError("Expecting a dict, but received an object of type {}."
                            .format(type(extra)))
        for param in self.params:
            # copy default params
            if param in self._defaultParamMap and to.hasParam(param.name):
                to._defaultParamMap[to.getParam(param.name)] = self._defaultParamMap[param]
            # copy explicitly set params
            if param in paramMap and to.hasParam(param.name):
                to._set(**{param.name: paramMap[param]})
        return to

    def _resetUid(self, newUid):
        """
        Changes the uid of this instance. This updates both
        the stored uid and the parent uid of params and param maps.
        This is used by persistence (loading).

        Parameters
        ----------
        newUid
            new uid to use, which is converted to unicode

        Returns
        -------
        :py:class:`Params`
            same instance, but with the uid and Param.parent values
            updated, including within param maps
        """
        newUid = str(newUid)
        self.uid = newUid
        newDefaultParamMap = dict()
        newParamMap = dict()
        for param in self.params:
            newParam = copy.copy(param)
            newParam.parent = newUid
            if param in self._defaultParamMap:
                newDefaultParamMap[newParam] = self._defaultParamMap[param]
            if param in self._paramMap:
                newParamMap[newParam] = self._paramMap[param]
            param.parent = newUid
        self._defaultParamMap = newDefaultParamMap
        self._paramMap = newParamMap
        return self



from dataclasses import dataclass
from typing import Optional
from typing import Union

import pendulum
from typeguard import typechecked

from tecton_core.errors import TectonValidationError


@dataclass
class BaseMaterializationContext:
    _feature_start_time_DONT_ACCESS_DIRECTLY: pendulum.DateTime
    _feature_end_time_DONT_ACCESS_DIRECTLY: pendulum.DateTime
    _batch_schedule_DONT_ACCESS_DIRECTLY: pendulum.Duration

    @property
    def start_time(self) -> pendulum.DateTime:
        return self._feature_start_time_DONT_ACCESS_DIRECTLY

    @property
    def end_time(self) -> pendulum.DateTime:
        return self._feature_end_time_DONT_ACCESS_DIRECTLY

    @property
    def batch_schedule(self) -> pendulum.Duration:
        return self._batch_schedule_DONT_ACCESS_DIRECTLY

    @typechecked
    def time_filter_sql(self, timestamp_expr: str) -> str:
        # Use atom string to include the timezone.
        return f"('{self.start_time.to_atom_string()}' <= ({timestamp_expr}) AND ({timestamp_expr}) < '{self.end_time.to_atom_string()}')"

    def time_filter_pyspark(self, timestamp_expr: Union[str, "pyspark.sql.Column"]) -> "pyspark.sql.Column":  # type: ignore
        from pyspark.sql.functions import expr
        from pyspark.sql.functions import lit

        if isinstance(timestamp_expr, str):
            timestamp_expr = expr(timestamp_expr)

        # Use atom string to include the timezone.
        return (lit(self.start_time.to_atom_string()) <= timestamp_expr) & (
            timestamp_expr < lit(self.end_time.to_atom_string())
        )

    def feature_time_filter_pyspark(self, timestamp_expr: Union[str, "pyspark.sql.Column"]) -> "pyspark.sql.Column":  # type: ignore
        return self.time_filter_pyspark(timestamp_expr)

    # EVERTHING BELOW IS DEPRECATED
    @property
    def feature_start_time(self) -> pendulum.DateTime:
        return self._feature_start_time_DONT_ACCESS_DIRECTLY

    @property
    def feature_end_time(self) -> pendulum.DateTime:
        return self._feature_end_time_DONT_ACCESS_DIRECTLY

    @property
    def feature_start_time_string(self) -> str:
        return self.feature_start_time.to_datetime_string()

    @property
    def feature_end_time_string(self) -> str:
        return self.feature_end_time.to_datetime_string()

    @typechecked
    def feature_time_filter_sql(self, timestamp_expr: str) -> str:
        return self.time_filter_sql(timestamp_expr)


@dataclass
class UnboundMaterializationContext(BaseMaterializationContext):
    """
    This is only meant for instantiation in transformation default args. Using it directly will fail.
    """

    @property
    def batch_schedule(self):
        raise TectonValidationError(
            "tecton.materialization_context() must be passed in via a kwarg default only. Instantiation in function body is not allowed."
        )

    @property
    def start_time(self):
        raise TectonValidationError(
            "tecton.materialization_context() must be passed in via a kwarg default only. Instantiation in function body is not allowed."
        )

    @property
    def end_time(self):
        raise TectonValidationError(
            "tecton.materialization_context() must be passed in via a kwarg default only. Instantiation in function body is not allowed."
        )

    @property
    def feature_start_time(self):
        raise TectonValidationError(
            "tecton.materialization_context() must be passed in via a kwarg default only. Instantiation in function body is not allowed."
        )

    @property
    def feature_end_time(self):
        raise TectonValidationError(
            "tecton.materialization_context() must be passed in via a kwarg default only. Instantiation in function body is not allowed."
        )


@dataclass
class BoundMaterializationContext(BaseMaterializationContext):
    @classmethod
    def create(cls, feature_start_time, feature_end_time):
        # user facing version
        return BoundMaterializationContext(
            _feature_start_time_DONT_ACCESS_DIRECTLY=feature_start_time,
            _feature_end_time_DONT_ACCESS_DIRECTLY=feature_end_time,
            # batch_schedule is passed by pipeline helper
            _batch_schedule_DONT_ACCESS_DIRECTLY=pendulum.duration(seconds=0),
        )

    @classmethod
    def _create_internal(cls, feature_start_time, feature_end_time, batch_schedule):
        # should only be used in pipeline_helper
        return BoundMaterializationContext(
            _feature_start_time_DONT_ACCESS_DIRECTLY=feature_start_time,
            _feature_end_time_DONT_ACCESS_DIRECTLY=feature_end_time,
            _batch_schedule_DONT_ACCESS_DIRECTLY=batch_schedule,
        )

    @classmethod
    def _create_from_period(cls, feature_time_limits: Optional[pendulum.Period], batch_schedule: pendulum.Duration):
        feature_start_time = (
            feature_time_limits.start
            if feature_time_limits is not None
            else pendulum.from_timestamp(0, pendulum.tz.UTC)
        )
        feature_end_time = feature_time_limits.end if feature_time_limits is not None else pendulum.datetime(2100, 1, 1)
        return BoundMaterializationContext(
            _feature_start_time_DONT_ACCESS_DIRECTLY=feature_start_time,
            _feature_end_time_DONT_ACCESS_DIRECTLY=feature_end_time,
            _batch_schedule_DONT_ACCESS_DIRECTLY=batch_schedule,
        )


def materialization_context():
    """
    Used as a default value for a Feature View or Transformation with a materialization context parameter.

    ``context.start_time`` and ``context.end_time`` return a :class:`datetime.datetime` object equal to the beginning and end of the period being materialized respectively. For example for a batch feature view materializing data from May 1st, 2022, ``context.start_time = datetime(2022, 5, 1)`` and ``context.end_time = datetime(2022, 5, 2)``.

    The datetimes can be used in SQL query strings directly (the datetime object will be cast to an atom-formatted timestamp string and inlined as a constant in the SQL query).

    Example usage:

    .. code-block:: python

        from tecton import batch_feature_view, materialization_context
        from datetime import datetime, timedelta

        @batch_feature_view(
            sources=[transactions],
            entities=[user],
            mode='spark_sql',
            batch_schedule=timedelta(days=1),
            feature_start_time=datetime(2020, 10, 10),
        )
        def user_last_transaction_amount(transactions, context=materialization_context()):
            return f'''
                SELECT
                    USER_ID,
                    AMOUNT,
                    TIMESTAMP
                FROM
                    {transactions}
                WHERE TIMESTAMP >= TO_TIMESTAMP("{context.start_time}") -- e.g. TO_TIMESTAMP("2022-05-01T00:00:00+00:00")
                    AND TIMESTAMP < TO_TIMESTAMP("{context.end_time}") -- e.g. TO_TIMESTAMP("2022-05-02T00:00:00+00:00")
                '''
    """
    dummy_time = pendulum.datetime(1970, 1, 1)
    dummy_period = pendulum.duration()
    return UnboundMaterializationContext(
        _feature_start_time_DONT_ACCESS_DIRECTLY=dummy_time,
        _feature_end_time_DONT_ACCESS_DIRECTLY=dummy_time,
        _batch_schedule_DONT_ACCESS_DIRECTLY=dummy_period,
    )



import tempfile
from datetime import datetime
from datetime import timedelta
from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

import pandas
import pendulum
from pyspark.sql import DataFrame
from pyspark.sql.streaming import StreamingQuery

import tecton
from tecton._internals import errors
from tecton._internals.rewrite import MockDataRewrite
from tecton.framework import data_frame
from tecton.run_api_consts import AGGREGATION_LEVEL_DISABLED
from tecton.run_api_consts import AGGREGATION_LEVEL_FULL
from tecton.run_api_consts import AGGREGATION_LEVEL_PARTIAL
from tecton.run_api_consts import DEFAULT_AGGREGATION_TILES_WINDOW_END_COLUMN_NAME
from tecton.run_api_consts import DEFAULT_AGGREGATION_TILES_WINDOW_START_COLUMN_NAME
from tecton.run_api_consts import SUPPORTED_AGGREGATION_LEVEL_VALUES
from tecton.tecton_context import TectonContext
from tecton_core import logger as logger_lib
from tecton_core import time_utils
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper as FeatureDefinition
from tecton_core.feature_definition_wrapper import FrameworkVersion
from tecton_core.query.builder import ANCHOR_TIME
from tecton_core.query.builder import build_get_full_agg_features
from tecton_core.query.builder import build_materialization_querytree
from tecton_core.query.builder import build_pipeline_querytree
from tecton_core.query.builder import WINDOW_END_COLUMN_NAME
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.nodes import AddAnchorTimeNode
from tecton_core.query.nodes import ConvertEpochToTimestampNode
from tecton_core.query.nodes import RenameColsNode
from tecton_core.query.nodes import UserSpecifiedDataNode
from tecton_core.query.sql_compat import default_case
from tecton_proto.args import pipeline_pb2
from tecton_proto.args.feature_view_pb2 import BackfillConfigMode
from tecton_proto.data import feature_view_pb2
from tecton_spark import materialization_plan
from tecton_spark.partial_aggregations import partial_aggregate_column_renames
from tecton_spark.pipeline_helper import get_all_input_ds_id_map
from tecton_spark.pipeline_helper import get_all_input_keys
from tecton_spark.pipeline_helper import run_mock_odfv_pipeline
from tecton_spark.spark_helper import check_spark_version

logger = logger_lib.get_logger("RunApi")


def maybe_warn_incorrect_time_range_size(
    fd: FeatureDefinition, start_time: datetime, end_time: datetime, aggregation_level: Optional[str]
):
    time_range = end_time - start_time
    if fd.is_temporal_aggregate:
        if fd.is_continuous:
            # There should not be any time range warnings for continuous aggregates.
            return
        slide_interval = fd.aggregate_slide_interval.ToTimedelta()
        if aggregation_level == AGGREGATION_LEVEL_FULL:
            max_aggregation = max((f.window.ToTimedelta() for f in fd.trailing_time_window_aggregation.features))
            if time_range < max_aggregation:
                logger.warning(
                    f"Run time range ({start_time}, {end_time}) is smaller than the maximum aggregation size: {max_aggregation}. This may lead to incorrect aggregate feature values."
                )
            if time_range.total_seconds() % slide_interval.total_seconds() != 0:
                logger.warning(
                    f"Run time range ({start_time}, {end_time}) is not a multiple of the aggregation_interval: {slide_interval}. This may lead to incorrect aggregate feature values, since Tecton pre-aggregates data in smaller time windows based on the aggregation_interval size."
                )
        elif aggregation_level == AGGREGATION_LEVEL_PARTIAL:
            if time_range.total_seconds() % slide_interval.total_seconds() != 0:
                logger.warning(
                    f"Run time range ({start_time}, {end_time}) is not a multiple of the aggregation_interval: {slide_interval}. This may lead to incorrect aggregate feature values, since Tecton pre-aggregates data in smaller time windows based on the aggregation_interval size."
                )
    elif fd.is_incremental_backfill and time_range != fd.batch_materialization_schedule:
        logger.warning(
            f"Run time range ({start_time}, {end_time}) is not equivalent to the batch_schedule: {fd.batch_materialization_schedule}. This may lead to incorrect feature values since feature views with incremental_backfills typically implicitly rely on the materialization range being equivalent to the batch_schedule."
        )


def run_batch(
    fd: FeatureDefinition,
    feature_start_time: datetime,
    feature_end_time: datetime,
    mock_inputs: Dict[str, Union[pandas.DataFrame, DataFrame]],
    framework_version: FrameworkVersion,
    aggregation_level: Optional[str],
) -> "tecton.framework.data_frame.TectonDataFrame":
    if not fd.is_on_demand:
        check_spark_version(fd.fv_spec.batch_cluster_config)

    return _querytree_run_batch(
        fd=fd,
        feature_start_time=feature_start_time,
        feature_end_time=feature_end_time,
        mock_inputs=mock_inputs,
        framework_version=framework_version,
        aggregation_level=aggregation_level,
    )


def _build_run_batch_querytree(
    fd: FeatureDefinition,
    feature_end_time: datetime,
    feature_time_limits_aligned: pendulum.Period,
    aggregation_level: Optional[str],
) -> NodeRef:
    """Build run_batch query tree

    This assumes that inputs are validated already and is general (should not
    handle mocking data). Using mock data is considered a query tree rewrite.

    Any extra querytree nodes in this function should simply be a display-level
    modification (like field rename, type change, etc).
    """
    if fd.is_temporal:
        qt = build_materialization_querytree(fd, for_stream=False, feature_data_time_limits=feature_time_limits_aligned)
        # For a BFV, the materialization querytree is an `AddAnchorTimeNode` wrapped around exactly what we want, so we
        # just extract the input node.
        assert isinstance(qt.node, AddAnchorTimeNode)
        return qt.node.inputs[0]
    elif fd.is_temporal_aggregate:
        if aggregation_level == AGGREGATION_LEVEL_PARTIAL:
            qt = build_materialization_querytree(
                fd,
                for_stream=False,
                feature_data_time_limits=feature_time_limits_aligned,
                include_window_end_time=True,
                aggregation_anchor_time=feature_end_time,
            )
            if fd.is_continuous:
                renames = {
                    **partial_aggregate_column_renames(
                        slide_interval_string=fd.get_aggregate_slide_interval_string,
                        trailing_time_window_aggregation=fd.trailing_time_window_aggregation,
                    ),
                }
                drop = [default_case(ANCHOR_TIME)]
            else:
                # The `PartialAggNode` returned by `build_materialization_querytree` converts timestamps to epochs. We convert back
                # from epochs to timestamps since timestamps are more readable.
                qt = ConvertEpochToTimestampNode(
                    qt, {col: fd.get_feature_store_format_version for col in (ANCHOR_TIME, WINDOW_END_COLUMN_NAME)}
                ).as_ref()
                renames = {
                    ANCHOR_TIME: DEFAULT_AGGREGATION_TILES_WINDOW_START_COLUMN_NAME,
                    WINDOW_END_COLUMN_NAME: DEFAULT_AGGREGATION_TILES_WINDOW_END_COLUMN_NAME,
                    **partial_aggregate_column_renames(
                        slide_interval_string=fd.get_aggregate_slide_interval_string,
                        trailing_time_window_aggregation=fd.trailing_time_window_aggregation,
                    ),
                }
                drop = []
            return RenameColsNode(qt, mapping=renames, drop=drop).as_ref()
        elif aggregation_level == AGGREGATION_LEVEL_DISABLED:
            return build_pipeline_querytree(fd, for_stream=False, feature_data_time_limits=feature_time_limits_aligned)
        elif aggregation_level == AGGREGATION_LEVEL_FULL:
            qt = build_get_full_agg_features(
                fd,
                from_source=True,
                feature_data_time_limits=feature_time_limits_aligned,
                aggregation_anchor_time=feature_end_time,
            )
            return qt

    raise Exception("Unsupported batch query tree")


def _querytree_run_batch(
    fd: FeatureDefinition,
    feature_start_time: datetime,
    feature_end_time: datetime,
    mock_inputs: Dict[str, Union[pandas.DataFrame, DataFrame]],
    framework_version: FrameworkVersion,
    aggregation_level: Optional[str],
) -> "tecton.framework.data_frame.DataFrame":
    aggregation_level = validate_and_get_aggregation_level(fd, aggregation_level)

    # Validate that mock_inputs' keys.
    validate_batch_mock_inputs_keys(mock_inputs, fd)

    feature_time_limits_aligned = pendulum.period(feature_start_time, feature_end_time)

    qt = _build_run_batch_querytree(fd, feature_end_time, feature_time_limits_aligned, aggregation_level)

    input_ds_id_map = get_all_input_ds_id_map(fd.pipeline.root)
    user_data_node_metadata = {"timestamp_key": fd.timestamp_key}
    mock_data_sources = {
        input_ds_id_map[key]: NodeRef(
            UserSpecifiedDataNode(data_frame.TectonDataFrame._create(mock_inputs[key]), user_data_node_metadata)
        )
        for key in mock_inputs.keys()
    }
    MockDataRewrite(mock_data_sources).rewrite(qt)

    return tecton.framework.data_frame.TectonDataFrame._create(qt)


def run_stream(fd: FeatureDefinition, output_temp_table: str) -> StreamingQuery:
    check_spark_version(fd.fv_spec.stream_cluster_config)
    plan = materialization_plan.get_stream_materialization_plan(
        spark=TectonContext.get_instance()._spark,
        feature_definition=fd,
    )
    spark_df = plan.online_store_data_frame
    with tempfile.TemporaryDirectory() as d:
        return (
            spark_df.writeStream.format("memory")
            .queryName(output_temp_table)
            .option("checkpointLocation", d)
            .outputMode("append")
            .start()
        )


def run_ondemand(
    fd: FeatureDefinition, fv_name: str, mock_inputs: Dict[str, Union[Dict[str, Any], pandas.DataFrame, DataFrame]]
) -> Union[Dict[str, Any], "tecton.framework.data_frame.TectonDataFrame"]:
    for key in mock_inputs:
        if isinstance(mock_inputs[key], DataFrame):
            mock_inputs[key] = mock_inputs[key].toPandas()

    # Validate that all the mock_inputs matchs with FV inputs, and that num rows match across all mock_inputs.
    validate_ondemand_mock_inputs_keys(mock_inputs, fd.pipeline)

    # Execute ODFV pipeline to get output DataFrame.
    output = run_mock_odfv_pipeline(
        pipeline=fd.pipeline,
        transformations=fd.transformations,
        name=fv_name,
        mock_inputs=mock_inputs,
    )
    if isinstance(output, pandas.DataFrame):
        output = tecton.framework.data_frame.TectonDataFrame._create(output)

    return output


def validate_and_get_aggregation_level(fd: FeatureDefinition, aggregation_level: Optional[str]) -> str:
    # Set default aggregation_level value.
    if aggregation_level is None:
        if fd.is_temporal_aggregate:
            aggregation_level = AGGREGATION_LEVEL_FULL
        else:
            aggregation_level = AGGREGATION_LEVEL_DISABLED

    if aggregation_level not in SUPPORTED_AGGREGATION_LEVEL_VALUES:
        raise errors.FV_INVALID_ARG_VALUE(
            "aggregation_level", str(aggregation_level), str(SUPPORTED_AGGREGATION_LEVEL_VALUES)
        )

    return aggregation_level


# For single-batch-schedule-interval-per-job backfill, validate the followings.
# - Only support single-tile run.
# - Don't allow passing `feature_start_time` without feature_end_time since it may be confusing that the tile time
#   range goes into the future.
def _validate_feature_time_for_backfill_config(
    fv_proto: feature_view_pb2.FeatureView,
    feature_start_time: Optional[Union[pendulum.DateTime, datetime]],
    feature_end_time: Optional[Union[pendulum.DateTime, datetime]],
    feature_time_limits_aligned: pendulum.Period,
):
    # TODO(raviphol): Use is_incremental_backfill once D9614 is landed.
    if not fv_proto.HasField("temporal"):
        return
    if not fv_proto.temporal.HasField("backfill_config"):
        return
    backfill_config_mode = fv_proto.temporal.backfill_config.mode
    if backfill_config_mode is not BackfillConfigMode.BACKFILL_CONFIG_MODE_SINGLE_BATCH_SCHEDULE_INTERVAL_PER_JOB:
        return

    if feature_start_time and not feature_end_time:
        raise errors.BFC_MODE_SINGLE_REQUIRED_FEATURE_END_TIME_WHEN_START_TIME_SET

    schedule_interval_seconds = fv_proto.materialization_params.schedule_interval.ToSeconds()
    if schedule_interval_seconds == 0:
        raise errors.INTERNAL_ERROR("Materialization schedule interval not found.")

    num_tile = feature_time_limits_aligned.in_seconds() // schedule_interval_seconds
    if num_tile > 1:
        raise errors.BFC_MODE_SINGLE_INVALID_FEATURE_TIME_RANGE


# Validate that mock_inputs keys are a subset of data sources.
def validate_batch_mock_inputs_keys(mock_inputs, fd: FeatureDefinition):
    expected_input_names = get_all_input_keys(fd.pipeline.root)
    mock_inputs_keys = set(mock_inputs.keys())
    if not mock_inputs_keys.issubset(expected_input_names):
        raise errors.FV_INVALID_MOCK_INPUTS(mock_inputs_keys, expected_input_names)


# Validate that mock_inputs keys are exact match with expected inputs.
def validate_ondemand_mock_inputs_keys(
    mock_inputs: Dict[str, Union[Dict[str, Any], pandas.DataFrame]],
    pipeline: pipeline_pb2.Pipeline,
):
    expected_input_names = get_all_input_keys(pipeline.root)
    mock_inputs_keys = set(mock_inputs.keys())
    if mock_inputs_keys != expected_input_names:
        raise errors.FV_INVALID_MOCK_INPUTS(mock_inputs_keys, expected_input_names)

    # Get num row for all FV mock_inputs with DF types, to validate that they match.
    input_df_row_counts = set()
    for input in mock_inputs.values():
        if isinstance(input, pandas.DataFrame):
            input_df_row_counts.add(len(input.index))
    if len(input_df_row_counts) > 1:
        raise errors.FV_INVALID_MOCK_INPUTS_NUM_ROWS(input_df_row_counts)


# Check that schema of each mock inputs matches with data sources.
def _validate_input_dataframe_schema(input_name, dataframe: DataFrame, spark_schema):
    columns = sorted(dataframe.columns)
    expected_column_names = sorted([field.name for field in spark_schema.fields])

    # Validate mock input's schema against expected schema.
    if not expected_column_names == columns:
        raise errors.FV_INVALID_MOCK_INPUT_SCHEMA(input_name, set(columns), set(expected_column_names))


def _get_ds_by_id(data_sources, id: str):
    for ds in data_sources:
        if ds.id == id:
            return ds
    return None


# Align feature start and end times with materialization schedule interval.
def _align_times(
    feature_start_time: datetime,
    feature_end_time: datetime,
    fv_proto: feature_view_pb2.FeatureView,
    schedule_interval: timedelta,
) -> pendulum.Period:
    # Align feature_end_time upward to the nearest materialization schedule interval.
    feature_end_time = time_utils.align_time_upwards(feature_end_time, schedule_interval)

    # Align feature_start_time downward to the nearest materialization schedule interval.
    feature_start_time = time_utils.align_time_downwards(feature_start_time, schedule_interval)
    return pendulum.period(feature_start_time, feature_end_time)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/column_type.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n%tecton_proto/common/column_type.proto\x12\x13tecton_proto.common*\xab\x02\n\nColumnType\x12\x17\n\x13\x43OLUMN_TYPE_UNKNOWN\x10\x00\x12\x15\n\x11\x43OLUMN_TYPE_INT64\x10\x01\x12\x16\n\x12\x43OLUMN_TYPE_DOUBLE\x10\x02\x12\x16\n\x12\x43OLUMN_TYPE_STRING\x10\x03\x12\x14\n\x10\x43OLUMN_TYPE_BOOL\x10\x06\x12\x1c\n\x18\x43OLUMN_TYPE_STRING_ARRAY\x10\x07\x12\x1b\n\x17\x43OLUMN_TYPE_INT64_ARRAY\x10\x08\x12\x1c\n\x18\x43OLUMN_TYPE_DOUBLE_ARRAY\x10\t\x12\x1b\n\x17\x43OLUMN_TYPE_FLOAT_ARRAY\x10\n\x12%\n!COLUMN_TYPE_DERIVE_FROM_DATA_TYPE\x10\x0b\"\x04\x08\x04\x10\x04\"\x04\x08\x05\x10\x05\x42?\n\x11\x63om.tecton.commonP\x01Z(github.com/tecton-ai/tecton_proto/common')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.common.column_type_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\021com.tecton.commonP\001Z(github.com/tecton-ai/tecton_proto/common'
  _COLUMNTYPE._serialized_start=63
  _COLUMNTYPE._serialized_end=362
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/spark_api/jobs.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.spark_common import clusters_pb2 as tecton__proto_dot_spark__common_dot_clusters__pb2
from tecton_proto.spark_common import libraries_pb2 as tecton__proto_dot_spark__common_dot_libraries__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!tecton_proto/spark_api/jobs.proto\x12\x16tecton_proto.spark_api\x1a(tecton_proto/spark_common/clusters.proto\x1a)tecton_proto/spark_common/libraries.proto\"\xb7\x03\n\x19PythonMaterializationTask\x12\x38\n\x18materialization_path_uri\x18\x01 \x01(\tR\x16materializationPathUri\x12n\n\x0f\x62\x61se_parameters\x18\x02 \x03(\x0b\x32\x45.tecton_proto.spark_api.PythonMaterializationTask.BaseParametersEntryR\x0e\x62\x61seParameters\x12V\n\x08taskType\x18\x03 \x01(\x0e\x32:.tecton_proto.spark_api.PythonMaterializationTask.TaskTypeR\x08taskType\x1a\x41\n\x13\x42\x61seParametersEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"U\n\x08TaskType\x12\r\n\tSTREAMING\x10\x00\x12\t\n\x05\x42\x41TCH\x10\x01\x12\n\n\x06INGEST\x10\x02\x12\x0c\n\x08\x44\x45LETION\x10\x03\x12\x15\n\x11\x44\x45LTA_MAINTENANCE\x10\x04\"\x98\x04\n\x0fStartJobRequest\x12H\n\x0bnew_cluster\x18\x01 \x01(\x0b\x32%.tecton_proto.spark_common.NewClusterH\x00R\nnewCluster\x12W\n\x10\x65xisting_cluster\x18\x07 \x01(\x0b\x32*.tecton_proto.spark_common.ExistingClusterH\x00R\x0f\x65xistingCluster\x12\x64\n\x14materialization_task\x18\x02 \x01(\x0b\x32\x31.tecton_proto.spark_api.PythonMaterializationTaskR\x13materializationTask\x12\x19\n\x08run_name\x18\x03 \x01(\tR\x07runName\x12@\n\tlibraries\x18\x04 \x03(\x0b\x32\".tecton_proto.spark_common.LibraryR\tlibraries\x12\'\n\x0ftimeout_seconds\x18\x05 \x01(\x05R\x0etimeoutSeconds\x12\x1f\n\x0bis_notebook\x18\x06 \x01(\x08R\nisNotebook\x12>\n\x1buse_stepped_materialization\x18\x0b \x01(\x08R\x19useSteppedMaterializationB\t\n\x07\x63lusterJ\x04\x08\t\x10\nJ\x04\x08\n\x10\x0b\")\n\x10StartJobResponse\x12\x15\n\x06run_id\x18\x01 \x01(\tR\x05runId\"&\n\rGetJobRequest\x12\x15\n\x06run_id\x18\x01 \x01(\tR\x05runId\"\x86\x03\n\x0eGetJobResponse\x12\x15\n\x06run_id\x18\x01 \x01(\tR\x05runId\x12\x15\n\x06job_id\x18\x07 \x01(\tR\x05jobId\x12 \n\x0crun_page_url\x18\x03 \x01(\tR\nrunPageUrl\x12(\n\x10spark_cluster_id\x18\x06 \x01(\tR\x0esparkClusterId\x12<\n\x07\x64\x65tails\x18\x04 \x01(\x0b\x32\".tecton_proto.spark_api.RunDetailsR\x07\x64\x65tails\x12o\n\x13\x61\x64\x64itional_metadata\x18\x05 \x03(\x0b\x32>.tecton_proto.spark_api.GetJobResponse.AdditionalMetadataEntryR\x12\x61\x64\x64itionalMetadata\x1a\x45\n\x17\x41\x64\x64itionalMetadataEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01J\x04\x08\x02\x10\x03\"\xd0\x01\n\nRunDetails\x12@\n\nrun_status\x18\x01 \x01(\x0e\x32!.tecton_proto.spark_api.RunStatusR\trunStatus\x12[\n\x12termination_reason\x18\x03 \x01(\x0e\x32,.tecton_proto.spark_api.RunTerminationReasonR\x11terminationReason\x12#\n\rstate_message\x18\x02 \x01(\tR\x0cstateMessage\"\'\n\x0eStopJobRequest\x12\x15\n\x06run_id\x18\x01 \x01(\tR\x05runId\"@\n\x0eListJobRequest\x12\x16\n\x06offset\x18\x01 \x01(\x05R\x06offset\x12\x16\n\x06marker\x18\x02 \x01(\tR\x06marker\"\x9f\x02\n\nRunSummary\x12\x15\n\x06run_id\x18\x01 \x01(\tR\x05runId\x12\x1b\n\trun_state\x18\x02 \x01(\tR\x08runState\x12)\n\x10resource_locator\x18\x03 \x01(\tR\x0fresourceLocator\x12k\n\x13\x61\x64\x64itional_metadata\x18\x04 \x03(\x0b\x32:.tecton_proto.spark_api.RunSummary.AdditionalMetadataEntryR\x12\x61\x64\x64itionalMetadata\x1a\x45\n\x17\x41\x64\x64itionalMetadataEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"|\n\x0fListJobResponse\x12\x36\n\x04runs\x18\x01 \x03(\x0b\x32\".tecton_proto.spark_api.RunSummaryR\x04runs\x12\x19\n\x08has_more\x18\x02 \x01(\x08R\x07hasMore\x12\x16\n\x06marker\x18\x03 \x01(\tR\x06marker*\xb6\x01\n\tRunStatus\x12\x16\n\x12RUN_STATUS_UNKNOWN\x10\x00\x12\x16\n\x12RUN_STATUS_PENDING\x10\x01\x12\x16\n\x12RUN_STATUS_RUNNING\x10\x02\x12\x16\n\x12RUN_STATUS_SUCCESS\x10\x03\x12\x14\n\x10RUN_STATUS_ERROR\x10\x04\x12\x1a\n\x16RUN_STATUS_TERMINATING\x10\x05\x12\x17\n\x13RUN_STATUS_CANCELED\x10\x06*\x98\x01\n\x14RunTerminationReason\x12\x1e\n\x1aUNKNOWN_TERMINATION_REASON\x10\x00\x12\x10\n\x0cJOB_FINISHED\x10\x01\x12\x16\n\x12MANUAL_CANCELATION\x10\x02\x12\x1f\n\x1bINSTANCE_ALLOCATION_FAILURE\x10\x03\x12\x15\n\x11NON_CLOUD_FAILURE\x10\x04\x42\x18\n\x14\x63om.tecton.spark_apiP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.spark_api.jobs_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\024com.tecton.spark_apiP\001'
  _PYTHONMATERIALIZATIONTASK_BASEPARAMETERSENTRY._options = None
  _PYTHONMATERIALIZATIONTASK_BASEPARAMETERSENTRY._serialized_options = b'8\001'
  _GETJOBRESPONSE_ADDITIONALMETADATAENTRY._options = None
  _GETJOBRESPONSE_ADDITIONALMETADATAENTRY._serialized_options = b'8\001'
  _RUNSUMMARY_ADDITIONALMETADATAENTRY._options = None
  _RUNSUMMARY_ADDITIONALMETADATAENTRY._serialized_options = b'8\001'
  _RUNSTATUS._serialized_start=2338
  _RUNSTATUS._serialized_end=2520
  _RUNTERMINATIONREASON._serialized_start=2523
  _RUNTERMINATIONREASON._serialized_end=2675
  _PYTHONMATERIALIZATIONTASK._serialized_start=147
  _PYTHONMATERIALIZATIONTASK._serialized_end=586
  _PYTHONMATERIALIZATIONTASK_BASEPARAMETERSENTRY._serialized_start=434
  _PYTHONMATERIALIZATIONTASK_BASEPARAMETERSENTRY._serialized_end=499
  _PYTHONMATERIALIZATIONTASK_TASKTYPE._serialized_start=501
  _PYTHONMATERIALIZATIONTASK_TASKTYPE._serialized_end=586
  _STARTJOBREQUEST._serialized_start=589
  _STARTJOBREQUEST._serialized_end=1125
  _STARTJOBRESPONSE._serialized_start=1127
  _STARTJOBRESPONSE._serialized_end=1168
  _GETJOBREQUEST._serialized_start=1170
  _GETJOBREQUEST._serialized_end=1208
  _GETJOBRESPONSE._serialized_start=1211
  _GETJOBRESPONSE._serialized_end=1601
  _GETJOBRESPONSE_ADDITIONALMETADATAENTRY._serialized_start=1526
  _GETJOBRESPONSE_ADDITIONALMETADATAENTRY._serialized_end=1595
  _RUNDETAILS._serialized_start=1604
  _RUNDETAILS._serialized_end=1812
  _STOPJOBREQUEST._serialized_start=1814
  _STOPJOBREQUEST._serialized_end=1853
  _LISTJOBREQUEST._serialized_start=1855
  _LISTJOBREQUEST._serialized_end=1919
  _RUNSUMMARY._serialized_start=1922
  _RUNSUMMARY._serialized_end=2209
  _RUNSUMMARY_ADDITIONALMETADATAENTRY._serialized_start=1526
  _RUNSUMMARY_ADDITIONALMETADATAENTRY._serialized_end=1595
  _LISTJOBRESPONSE._serialized_start=2211
  _LISTJOBRESPONSE._serialized_end=2335
# @@protoc_insertion_point(module_scope)
